"""Module for classes that format variables."""

# Standard imports
from time import time
import socket
import re

# pattoo imports
from pattoo_shared import data
from pattoo_shared import files
from pattoo_shared import network
from pattoo_shared.constants import (
    DATA_INT, DATA_FLOAT, DATA_COUNT64, DATA_COUNT, DATA_STRING, DATA_NONE,
    DATAPOINT_KEYS, AGENT_METADATA_KEYS)


class Metadata():
    """Metadata related to a DataPoint."""

    def __init__(self, key, value):
        """Initialize the class.

        Args:
            key: Metadata key
            value: Metadata value

        Returns:
            None

        """
        # Initialize variables
        self.key = key
        self.value = value
        self.update_checksum = True

    def __repr__(self):
        """Return a representation of the attributes of the class.

        Args:
            None

        Returns:
            result: String representation.

        """
        # Create a printable variation of the value
        result = ('<{0} key={1}, value={2}>'.format(
            self.__class__.__name__, repr(self.key), repr(self.value))
        )
        return result


class ConverterMetadata(Metadata):
    """Metadata related to a Converter DataPoint."""

    def __init__(self, key, value, update_checksum=True):
        """Initialize the class.

        Args:
            key: Metadata key
            value: Metadata value
            update_checksum: When True, the Datapoint object will update its
                checksum using the metadata values of objects created with this
                class. The DataPoint checksums are used to create new database
                entries. Not all metadata changes should trigger the creation
                of a new datapoint for charting. The updating of an operating
                system kernel version as metadata may not necessarily mean a
                completely new data source.

        Returns:
            None

        Variables:
            self.key: Metadata key
            self.value: Metadata value
            self.valid: True if valid

        """
        # Inherit object
        Metadata.__init__(self, key, value)

        # Initialize variables
        self.update_checksum = bool(update_checksum)
        (self.key, self.value, self.valid) = _key_value_valid(
            key, value, metadata=True, override=True)

        # Reset valid
        self.valid = False not in [
            key in AGENT_METADATA_KEYS,
            self.valid
        ]


class DataPointMetadata(Metadata):
    """Metadata related to a Regular DataPoint."""

    def __init__(self, key, value, update_checksum=True):
        """Initialize the class.

        Args:
            key: Metadata key
            value: Metadata value
            update_checksum: When True, the Datapoint object will update its
                checksum using the metadata values of objects created with this
                class. The DataPoint checksums are used to create new database
                entries. Not all metadata changes should trigger the creation
                of a new datapoint for charting. The updating of an operating
                system kernel version as metadata may not necessarily mean a
                completely new data source.

        Returns:
            None

        Variables:
            self.key: Metadata key
            self.value: Metadata value
            self.valid: True if valid

        """
        # Inherit object
        Metadata.__init__(self, key, value)
        self.update_checksum = bool(update_checksum)

        # Initialize variables
        (self.key, self.value, self.valid) = _key_value_valid(
            key, value, metadata=True)


class DataPoint():
    """Variable representation for data retreived from a target.

    Stores individual datapoints polled by pattoo agents

    """

    def __init__(self, key, value, data_type=DATA_INT, timestamp=None):
        """Initialize the class.

        Args:
            key: Key related to data value
            value: Data value
            data_type: This MUST be one of the types listed in constants.py
            timestamp: Integer EPOCH timestamp in milliseconds

        Returns:
            None

        Variables:
            self.timestamp: Integer of epoch milliseconds
            self.valid: True if the object has a valid data_type
            self.checksum: Hash of self.key, self.data_type and metadata to
                ensure uniqueness when assigned to a target.

        """
        # Initialize variables
        (self.key, self.value, self.valid) = _key_value_valid(
            key, value, metadata=False)
        self.data_type = data_type
        self.metadata = {}
        self._metakeys = []

        # Round timestamp to the nearest millisecond.
        if data.is_numeric(timestamp) is False:
            self.timestamp = int(round(time(), 3) * 1000)
        else:
            self.timestamp = int(timestamp)

        # False validity if value is not of the right type
        self.valid = False not in [
            data_type in [
                DATA_INT,
                DATA_FLOAT,
                DATA_COUNT64,
                DATA_COUNT,
                DATA_STRING,
                DATA_NONE
            ],
            data_type is not False,
            data_type is not True,
            data_type is not None,
            self.valid is True
        ]

        # Validity check: Make sure numeric data_types have numeric values
        if False not in [
                data_type in [
                    DATA_INT,
                    DATA_FLOAT,
                    DATA_COUNT64,
                    DATA_COUNT
                ],
                self.valid is True,
                data.is_numeric(value) is False]:
            self.valid = False

        # Convert floatable strings to float, and integers to ints
        if False not in [
                self.valid is True,
                data.is_numeric(value) is True,
                isinstance(value, str) is True]:
            if data_type in [DATA_FLOAT, DATA_COUNT64, DATA_COUNT]:
                self.value = float(value)
            elif data_type in [DATA_INT]:
                self.value = int(float(value))

        # Convert strings to string
        if data_type in [DATA_STRING]:
            self.value = str(value)

        # Create checksum
        self.checksum = data.hashstring(
            '{}{}'.format(self.key, self.data_type))

    def __repr__(self):
        """Return a representation of the attributes of the class.

        Args:
            None

        Returns:
            result: String representation.

        """
        # Create a printable variation of the value
        printable_value = _strip_non_printable(self.key)
        result = ('''\
<{} key={}, value={}, data_type={}, timestamp={}, valid={}>\
'''.format(self.__class__.__name__,
           repr(printable_value), repr(self.value),
           repr(self.data_type), repr(self.timestamp),
           repr(self.valid))
        )
        return result

    def add(self, items):
        """Add DataPointMetadata to the internal self.metadata list.

        Args:
            items: A DataPointMetadata object list

        Returns:
            None

        """
        # Ensure there is a list of objects
        if isinstance(items, list) is False:
            items = [items]

        # Only append approved data types
        for item in items:
            if isinstance(item, Metadata) is True:
                # Ignore invalid values
                if item.valid is False or item.key in DATAPOINT_KEYS:
                    continue

                # Process
                if item.key not in self._metakeys:
                    self.metadata[item.key] = item.value
                    self._metakeys.append(item.key)
                    if bool(item.update_checksum) is True:
                        self.checksum = data.hashstring('''\
{}{}{}'''.format(self.checksum, item.key, item.value))


class PostingDataPoints():
    """Object defining DataPoint objects to post to the pattoo server."""

    def __init__(self, agent_id, polling_interval, datapoints):
        """Initialize the class.

        Args:
            agent_id: Unique ID of agent posting data
            polling_interval: Periodic interval over which the data was polled.
            datapoints: List of DataPoint objects

        Returns:
            None

        """
        # Initialize key variables
        self.pattoo_agent_id = agent_id
        self.pattoo_agent_polling_interval = polling_interval
        self.pattoo_datapoints = datapoints
        self.pattoo_timestamp = int(time() * 1000)

        # Validation tests
        self.valid = False not in [
            isinstance(self.pattoo_agent_id, str),
            isinstance(self.pattoo_agent_polling_interval, int),
            isinstance(self.pattoo_datapoints, dict),
            self.pattoo_agent_polling_interval is not False,
            self.pattoo_agent_polling_interval is not True,
        ]
        if self.valid is True:
            self.valid = False not in [
                self.valid,
                'datapoint_pairs' in datapoints.keys(),
                'key_value_pairs' in datapoints.keys()
                ]
        if self.valid is True:
            self.valid = False not in [
                self.valid,
                isinstance(datapoints['datapoint_pairs'], list),
                isinstance(datapoints['key_value_pairs'], dict)
                ]


class TargetDataPoints():
    """Object defining a list of DataPoint objects.

    Stores DataPoints polled from a specific ip_target.

    """

    def __init__(self, target):
        """Initialize the class.

        Args:
            target: Target polled to get the DataPoint objects

        Returns:
            None

        Variables:
            self.data: List of DataPoints retrieved from the target
            self.valid: True if the object is populated with DataPoints

        """
        # Initialize key variables
        self.data = []
        self.target = target
        self.valid = False
        self._checksums = []

    def __repr__(self):
        """Return a representation of the attributes of the class.

        Args:
            None

        Returns:
            result: String representation.

        """
        # Create a printable variation of the value
        result = (
            '<{0} target={1}, valid={2}, data={3}'
            ''.format(
                self.__class__.__name__,
                repr(self.target), repr(self.valid), repr(self.data)
            )
        )
        return result

    def add(self, items):
        """Append DataPoint to the internal self.data list.

        Args:
            items: A DataPoint object list

        Returns:
            None

        """
        # Ensure there is a list of objects
        if isinstance(items, list) is False:
            items = [items]

        # Only add DataPoint objects that are not duplicated
        for item in items:
            if isinstance(item, DataPoint) is True:
                if item.checksum not in self._checksums:
                    self.data.append(item)
                    self._checksums.append(item.checksum)

                # Set object as being.valid
                self.valid = False not in [bool(self.data), bool(self.target)]


class AgentPolledData():
    """Object defining data received from / sent by Agent.

    Only AgentPolledData objects can be submitted to the pattoo server through
    phttp.Post()

    """

    def __init__(self, agent_program, polling_interval):
        """Initialize the class.

        Args:
            agent_program: Name of agent program collecting the data
            polling_interval: Polling interval

        Returns:
            None

        Variables:
            self.data: List of TargetDataPoints objects created by polling
            self.valid: True if the object contains TargetDataPoints objects

        """
        # Initialize key variables
        self.agent_program = agent_program
        self.agent_hostname = socket.getfqdn()
        self.agent_timestamp = int(time() * 1000)
        self.data = []
        self.valid = False
        self.agent_polling_interval = polling_interval * 1000

        # Get the agent_id
        from .configuration import Config
        config = Config()
        self.agent_id = files.get_agent_id(agent_program, config)

    def __repr__(self):
        """Return a representation of the attributes of the class.

        Args:
            None

        Returns:
            result: String representation.

        """
        # Return
        result = ('''\
<{0} agent_id={1} agent_program={2}, agent_hostname={3}, timestamp={4} \
polling_interval={5}, valid={6}>\
'''.format(self.__class__.__name__, repr(self.agent_id),
           repr(self.agent_program), repr(self.agent_hostname),
           repr(self.agent_timestamp), repr(self.agent_polling_interval),
           repr(self.valid)))
        return result

    def add(self, items):
        """Append TargetDataPoints to the internal self.data list.

        Args:
            items: A TargetDataPoints object list

        Returns:
            None

        """
        # Do nothing if not a list
        if isinstance(items, list) is False:
            items = [items]

        # Only append approved data types
        for item in items:
            # Only append approved data types
            if isinstance(item, TargetDataPoints) is True:
                # Ignore invalid values
                if item.valid is False:
                    continue

                # Process
                self.data.append(item)

                # Set object as being.valid
                self.valid = False not in [
                    bool(self.agent_id),
                    bool(self.agent_program),
                    bool(self.agent_hostname),
                    bool(self.agent_polling_interval),
                    bool(self.data)]


class AgentAPIVariable():
    """Variable representation for data required by the AgentAPI."""

    def __init__(self, ip_bind_port=20201, ip_listen_address='0.0.0.0'):
        """Initialize the class.

        Args:
            ip_bind_port: ip_bind_port
            listen_address: TCP/IP address on which the API is listening.

        Returns:
            None

        """
        # Initialize variables
        self.ip_bind_port = ip_bind_port
        self.ip_listen_address = ip_listen_address

    def __repr__(self):
        """Return a representation of the attributes of the class.

        Args:
            None

        Returns:
            result: String representation.

        """
        result = ('''\
<{0} ip_bind_port={1}, ip_listen_address={2}>\
'''.format(self.__class__.__name__,
           repr(self.ip_bind_port),
           repr(self.ip_listen_address)
           )
        )
        return result


class PollingPoint():
    """Object used to track data to be polled."""

    def __init__(self, address=None, multiplier=1):
        """Initialize the class.

        Args:
            address: Address to poll
            multiplier: Multiplier to use when polled

        Returns:
            None

        """
        # Initialize variables
        self.address = address
        if data.is_numeric(multiplier) is True:
            self.multiplier = multiplier
        else:
            self.multiplier = 1
        self.valid = address is not None

        # Create checksum
        seed = '{}{}'.format(address, multiplier)
        self.checksum = data.hashstring(seed)

    def __repr__(self):
        """Return a representation of the attributes of the class.

        Args:
            None

        Returns:
            result: String representation.

        """
        result = ('''\
<{0} address={1}, multiplier={2}>\
'''.format(self.__class__.__name__,
           repr(self.address),
           repr(self.multiplier)
           ))
        return result


class TargetPollingPoints():
    """Object defining a list of PollingPoint objects.

    Stores PollingPoints polled from a specific ip_target.

    """

    def __init__(self, target):
        """Initialize the class.

        Args:
            target: Target polled to get the PollingPoint objects

        Returns:
            None

        Variables:
            self.data: List of PollingPoints retrieved from the target
            self.target: Name of target from which the data was received
            self.valid: True if the object is populated with PollingPoints

        """
        # Initialize key variables
        self.data = []
        self.target = target
        self.valid = False
        self._checksums = []

    def __repr__(self):
        """Return a representation of the attributes of the class.

        Args:
            None

        Returns:
            result: String representation.

        """
        # Create a printable variation of the value
        result = (
            '<{0} target={1}, valid={2}, data={3}>'
            ''.format(
                self.__class__.__name__,
                repr(self.target), repr(self.valid), repr(self.data)
            )
        )
        return result

    def add(self, items):
        """Append PollingPoint to the internal self.data list.

        Args:
            items: A PollingPoint object list

        Returns:
            None

        """
        # Ensure there is a list of objects
        if isinstance(items, list) is False:
            items = [items]

        # Only add PollingPoint objects that are not duplicated
        for item in items:
            if isinstance(item, PollingPoint) is True:
                # Ignore invalid values
                if item.valid is False:
                    continue

                # Add data to the list
                if item.checksum not in self._checksums:
                    self.data.append(item)

                # Set object as being.valid
                self.valid = False not in [bool(self.data), bool(self.target)]


class IPTargetPollingPoints(TargetPollingPoints):
    """Object defining a list of PollingPoint objects.

    Stores PollingPoints polled from a specific ip_target.

    """

    def __init__(self, target):
        """Initialize the class.

        Args:
            target: Target polled to get the PollingPoint objects

        Returns:
            None

        Variables:
            self.data: List of PollingPoints retrieved from the target
            self.target: Name of target from which the data was received
            self.valid: True if the object is populated with PollingPoints

        """
        # Inherit object
        TargetPollingPoints.__init__(self, target)

    def add(self, items):
        """Append PollingPoint to the internal self.data list.

        Args:
            items: A PollingPoint object list

        Returns:
            None

        """
        # Ensure there is a list of objects
        if isinstance(items, list) is False:
            items = [items]

        # Only add PollingPoint objects that are not duplicated
        for item in items:
            if isinstance(item, PollingPoint) is True:
                # Ignore invalid values
                if item.valid is False:
                    continue

                # Add data to the list
                if item.checksum not in self._checksums:
                    self.data.append(item)

                # Set object as being.valid
                self.valid = False not in [
                    bool(self.data), bool(network.get_ipaddress(self.target))]


def _strip_non_printable(value):
    """Strip non printable characters.

    Removes any non-printable characters and adds an indicator to the string
    when binary characters are found.

    Args:
        value: the value that you wish to strip

    Returns:
        printable_value: Printable string

    """
    # Initialize key variables
    printable_value = ''

    if isinstance(value, str) is False:
        printable_value = value
    else:
        # Filter all non-printable characters
        # (note that we must use join to account for the fact that Python 3
        # returns a generator)
        printable_value = ''.join(
            [x for x in value if x.isprintable() is True])
        if printable_value != value:
            if bool(printable_value) is True:
                printable_value = '{} '.format(printable_value)
            printable_value = '{}(contains binary)'.format(printable_value)

    # Return
    return printable_value


def _key_value_valid(key, value, metadata=False, override=False):
    """Create a standardized version of key, value.

    Args:
        key: Key
        value: Value
        metadata: If True, convert value to string and strip. Used for
            metadata key-value pairs.
        override: Allow the 'pattoo' string in the key. Used when posting
            DataPoint objects to the pattoo server.

    Returns:
        result: Tuple of (key, value, valid)

    """
    # Set variables
    valid = False not in [
        isinstance(key, (str, int, float)) is True,
        key is not True,
        key is not False,
        key is not None,
        isinstance(value, (str, int, float)) is True,
        value is not True,
        value is not False,
        value is not None,
        ]

    # Assign key, value
    if valid is True:
        key = str(key).lower().strip()

        # Reevaluate valid
        valid = False not in [
            valid,
            key != '']

        # 'pattoo' normally cannot be in the key.
        if override is False:
            valid = False not in [
                valid,
                'pattoo' not in key.lower()]

    # Assign values
    if valid is True:
        if bool(metadata) is True:
            value = str(value).strip()
    else:
        key = None
        value = None

    # Return
    result = (key, value, valid)
    return result


def _strip_pattoo(key):
    """Remove the string 'pattoo' from key.

    1) Remove the string 'pattoo' from key
    2) Replace dashes ('-') with underscores ('_')
    3) Remove leading underscores
    4) Strip leading and trailing white space
    5) Make the key lowercase

    Args:
        key: Key to process

    Returns:
        result: Key without string.

    """
    # Return
    result = str(key).lower().strip().replace('pattoo', '').replace('-', '_')
    result = re.sub(r'^(_)+(.*?)$', '\\2', result)
    return result



#!/usr/bin/env python3
"""Test the files module."""

# Standard imports
import unittest
import os
import tempfile
import sys
import json
from math import pi
from random import randint
import shutil
from random import random

# PIP imports
import yaml

# Try to create a working PYTHONPATH
EXEC_DIR = os.path.dirname(os.path.realpath(__file__))
ROOT_DIR = os.path.abspath(os.path.join(
    os.path.abspath(os.path.join(EXEC_DIR, os.pardir)), os.pardir))
_EXPECTED = '{0}pattoo-shared{0}tests{0}pattoo_shared_'.format(os.sep)
if EXEC_DIR.endswith(_EXPECTED) is True:
    # We need to prepend the path in case PattooShared has been installed
    # elsewhere on the system using PIP. This could corrupt expected results
    sys.path.insert(0, ROOT_DIR)
else:
    print('''This script is not installed in the "{0}" directory. Please fix.\
'''.format(_EXPECTED))
    sys.exit(2)

# Pattoo imports
from pattoo_shared import files
from pattoo_shared.configuration import Config
from tests.libraries.configuration import UnittestConfig


class TestBasicFunctions(unittest.TestCase):
    """Checks all functions and methods."""

    #########################################################################
    # General object setup
    #########################################################################

    prefix = 'unittest'
    agent_hostname = 'pattoo_host'
    config = Config()

    def test_execute(self):
        """Testing method or function named execute."""
        # Test known bad command
        command = '{}'.format(random())
        with self.assertRaises(SystemExit):
            files.execute(command)

        # Test known bad command, but don't die
        command = '{}'.format(random())
        result = files.execute(command, die=False)
        self.assertTrue(bool(result))

        # Test known good command (Works in Windows and Linux)
        result = files.execute('date')
        self.assertFalse(bool(result))

    def test_read_yaml_files(self):
        """Testing method or function named read_yaml_files."""
        # Initializing key variables
        dict_1 = {
            'key1': 1,
            'key2': 2,
            'key3': 3,
            'key4': 4,
        }

        dict_2 = {
            'key6': 6,
            'key7': 7,
        }
        dict_3 = {}

        # Populate a third dictionary with contents of other dictionaries.
        for key, value in dict_1.items():
            dict_3[key] = value

        for key, value in dict_2.items():
            dict_3[key] = value

        # Create temp file with known data
        directory = tempfile.mkdtemp()
        filenames = {
            '{}{}file_1.yaml'.format(directory, os.sep): dict_1,
            '{}{}file_2.yaml'.format(directory, os.sep): dict_2
        }
        for filename, data_dict in filenames.items():
            with open(filename, 'w') as fh_:
                yaml.safe_dump(data_dict, stream=fh_, default_flow_style=False)

        # Get Results
        result = files.read_yaml_files(directory)

        # Clean up
        for key in result.keys():
            self.assertEqual(dict_3[key], result[key])
        filelist = [
            next_file for next_file in os.listdir(
                directory) if next_file.endswith('.yaml')]
        for delete_file in filelist:
            delete_path = '{}{}{}'.format(directory, os.sep, delete_file)
            os.remove(delete_path)
        os.removedirs(directory)

    def test_read_yaml_file(self):
        """Testing function read_yaml_file."""
        # Initializing key variables
        dict_1 = {
            'key1': 1,
            'key2': 2,
            'key3': 3,
            'key4': 4,
        }

        # Create a temporary file without a json extension and test
        tmp = tempfile.NamedTemporaryFile(delete=False)
        with self.assertRaises(SystemExit):
            _ = files.read_yaml_file(tmp.name)
        os.remove(tmp.name)

        # Create temp file with known data
        directory = tempfile.mkdtemp()
        file_data = [
            (('{}{}file_1.yaml').format(directory, os.sep), dict_1)
        ]
        for item in file_data:
            filename = item[0]
            data_dict = item[1]
            with open(filename, 'w') as fh_:
                yaml.safe_dump(data_dict, stream=fh_, default_flow_style=False)

            # Get Results
            result = files.read_yaml_file(filename)

            # Test equivalence
            for key in result.keys():
                self.assertEqual(data_dict[key], result[key])

        # Clean up
        filelist = [
            next_file for next_file in os.listdir(
                directory) if next_file.endswith('.yaml')]
        for delete_file in filelist:
            delete_path = ('{}{}{}').format(directory, os.sep, delete_file)
            os.remove(delete_path)
        os.removedirs(directory)

    def test_read_json_files(self):
        """Testing method or function named read_json_files."""
        # Initializing key variables
        dict_1 = {
            'key1': 1,
            'key2': 2,
            'key3': 3,
            'key4': 4,
        }

        dict_2 = {
            'key6': 6,
            'key7': 7,
        }

        # Create a temporary file without a json extension and test
        directory = tempfile.mkdtemp()
        with self.assertRaises(SystemExit):
            _ = files.read_json_files(directory)
        os.removedirs(directory)

        # Create a temporary file without a json extension and test
        directory = tempfile.mkdtemp()
        with self.assertRaises(SystemExit):
            _ = files.read_json_files(directory, die=True)
        os.removedirs(directory)

        # Test with die being False. Nothing should happen
        directory = tempfile.mkdtemp()
        _ = files.read_json_files(directory, die=False)
        os.removedirs(directory)

        # Create temp file with known data
        directory = tempfile.mkdtemp()
        filenames = {
            '{}{}file_1.json'.format(directory, os.sep): dict_1,
            '{}{}file_2.json'.format(directory, os.sep): dict_2
        }
        for filename, data_dict in filenames.items():
            with open(filename, 'w') as filehandle:
                json.dump(data_dict, filehandle)

        # Get Results
        result = files.read_json_files(directory)

        # First test, only 2 files
        self.assertEqual(len(result), 2)

        # Clean up
        for filepath, data in result:
            self.assertEqual(filepath in filenames, True)
            for key, value in sorted(data.items()):
                self.assertEqual(filenames[filepath][key], value)
            os.remove(filepath)
        os.removedirs(directory)

    def test_read_json_file(self):
        """Testing function read_json_file."""
        # Initialize key variables
        data = {
            'key1': 1,
            'key2': 2,
            'key3': 3,
            'key4': 4,
        }

        # Create a temporary file without a json extension and test
        tmp = tempfile.NamedTemporaryFile(delete=False)
        with self.assertRaises(SystemExit):
            _ = files.read_json_file(tmp.name)
        os.remove(tmp.name)

        # Create a temporary file without a json extension and test
        tmp = tempfile.NamedTemporaryFile(delete=False)
        with self.assertRaises(SystemExit):
            _ = files.read_json_file(tmp.name, die=True)
        os.remove(tmp.name)

        # Test with die being False. Nothing should happen
        tmp = tempfile.NamedTemporaryFile(delete=False)
        _ = files.read_json_file(tmp.name, die=False)
        os.remove(tmp.name)

        # Create json file and test
        tmp = tempfile.NamedTemporaryFile(delete=False, suffix='.json')
        with open(tmp.name, 'w') as f_handle:
            json.dump(data, f_handle)
        result = files.read_json_file(tmp.name)
        self.assertEqual(len(result), 4)
        for key, value in sorted(result.items()):
            self.assertEqual(data[key], value)
        os.remove(tmp.name)

    def test_mkdir(self):
        """Testing function mkdir."""
        # Test with file, not directory
        tmpfile = tempfile.NamedTemporaryFile(delete=False).name
        open(tmpfile, 'a').close()
        with self.assertRaises(SystemExit):
            files.mkdir(tmpfile)
        os.remove(tmpfile)

        # Create a sub directory of a temp directory.
        directory = '{0}tmp{0}test_pattoo-unittest{0}{1}.fake'.format(
            os.sep, randint(1, 10000) * pi)
        files.mkdir(directory)
        self.assertTrue(os.path.isdir(directory))
        shutil.rmtree(directory)

    def test_pid_file(self):
        """Testing function pid_file."""
        # Test
        filename = files._File(self.config)
        expected = filename.pid(self.prefix)
        result = files.pid_file(self.prefix, self.config)
        self.assertEqual(result, expected)

    def test_lock_file(self):
        """Testing function lock_file."""
        # Test
        filename = files._File(self.config)
        expected = filename.lock(self.prefix)
        result = files.lock_file(self.prefix, self.config)
        self.assertEqual(result, expected)

    def test_agent_id_file(self):
        """Testing function agent_id_file."""
        # Test
        filename = files._File(self.config)
        expected = filename.agent_id(self.prefix)
        result = files.agent_id_file(self.prefix, self.config)
        self.assertEqual(result, expected)

    def test_get_agent_id(self):
        """Testing method or function named get_agent_id."""
        # Test. Agent_id shouldn't change
        agent_name = random()
        expected = files.get_agent_id(agent_name, self.config)
        result = files.get_agent_id(agent_name, self.config)
        self.assertEqual(result, expected)

        # Delete the file, make sure a new ID is generated
        filename = files.agent_id_file(agent_name, self.config)
        os.remove(filename)
        _expected = files.get_agent_id(agent_name, self.config)
        _result = files.get_agent_id(agent_name, self.config)
        self.assertEqual(_result, _expected)
        self.assertNotEqual(expected, _expected)

    def test__generate_agent_id(self):
        """Testing method or function named _generate_agent_id."""
        pass


class Test_Directory(unittest.TestCase):
    """Checks all functions and methods."""

    #########################################################################
    # General object setup
    #########################################################################

    config = Config()

    def test___init__(self):
        """Testing function __init__."""
        # Test
        directory = files._Directory(self.config)
        expected = self.config.daemon_directory()
        result = directory._root
        self.assertEqual(result, expected)

    def test_pid(self):
        """Testing function pid."""
        # Test
        directory = files._Directory(self.config)
        expected = '{}{}pid'.format(self.config.system_daemon_directory(), os.sep)
        result = directory.pid()
        self.assertEqual(result, expected)

    def test_lock(self):
        """Testing function lock."""
        # Test
        directory = files._Directory(self.config)
        expected = '{}{}lock'.format(self.config.system_daemon_directory(), os.sep)
        result = directory.lock()
        self.assertEqual(result, expected)

    def test_agent_id(self):
        """Testing function agent_id."""
        # Test
        directory = files._Directory(self.config)
        expected = '{}{}agent_id'\
                   .format(self.config.daemon_directory(), os.sep)
        result = directory.agent_id()
        self.assertEqual(result, expected)


class Test_File(unittest.TestCase):
    """Checks all functions and methods."""

    #########################################################################
    # General object setup
    #########################################################################

    prefix = 'unittest'
    agent_hostname = 'pattoo_host'
    config = Config()

    def test___init__(self):
        """Testing function __init__."""
        pass

    def test_pid(self):
        """Testing function pid."""
        # Test
        filename = files._File(self.config)
        result = filename.pid(self.prefix)
        self.assertTrue(os.path.isdir(os.path.dirname(result)))

    def test_lock(self):
        """Testing function lock."""
        # Test
        filename = files._File(self.config)
        result = filename.lock(self.prefix)
        self.assertTrue(os.path.isdir(os.path.dirname(result)))

    def test_agent_id(self):
        """Testing function agent_id."""
        filename = files._File(self.config)
        result = filename.agent_id(self.prefix)
        self.assertTrue(os.path.isdir(os.path.dirname(result)))


if __name__ == '__main__':
    # Make sure the environment is OK to run unittests
    UnittestConfig().create()

    # Do the unit test
    unittest.main()



#!/usr/bin/env python3
"""Test pattoo shared packages script."""
import os
import unittest
from random import random
import sys
import tempfile
from collections import namedtuple

# Try to create a working PYTHONPATH
EXEC_DIR = os.path.dirname(os.path.realpath(__file__))
ROOT_DIR = os.path.abspath(os.path.join(
    os.path.abspath(os.path.join(
        os.path.abspath(os.path.join(
            EXEC_DIR, os.pardir)), os.pardir)), os.pardir))
_EXPECTED = '''\
{0}pattoo-shared{0}tests{0}pattoo_shared_{0}installation'''.format(os.sep)
if EXEC_DIR.endswith(_EXPECTED) is True:
    # We need to prepend the path in case PattooShared has been installed
    # elsewhere on the system using PIP. This could corrupt expected results
    sys.path.insert(0, ROOT_DIR)
else:
    print('''This script is not installed in the "{0}" directory. Please fix.\
'''.format(_EXPECTED))
    sys.exit(2)

# Pattoo imports
from tests.libraries.configuration import UnittestConfig
from pattoo_shared import data
from pattoo_shared.installation import shared, environment
from pattoo_shared.installation import packages


class TestPackages(unittest.TestCase):
    """Checks all functions for the Pattoo packages script."""

    def setUp(self):
        """Prepare each test for testing."""
        self.venv_dir = tempfile.mkdtemp()
        self.venv = environment.environment_setup(self.venv_dir)

    def tearDown(self):
        """Cleanup each test after testing."""
        self.venv.deactivate()

    def test_install_package(self):
        """Unittest to test the install_package function."""
        # Test with expected behaviour
        with self.subTest():
            # Attempt to install a test package
            packages.install_package('tweepy', verbose=False)

            # Try except to determine if package was installed
            try:
                import tweepy
                result = True
            except ModuleNotFoundError:
                result = False
            self.assertTrue(result)

        # Test case that would cause the install_package
        # function to fail
        with self.subTest():
            with self.assertRaises(SystemExit) as cm_:
                packages.install_package('This does not exist', False)
            self.assertEqual(cm_.exception.code, 2)

        # Test with outdated package version
        with self.subTest():
            packages.install_package('matplotlib==3.3.0', False)
            expected = '3.3.0'
            result = packages.installed_version('matplotlib')
            self.assertEqual(result, expected)

        # Try reinstalling the previous version
        with self.subTest():
            packages.install_package('matplotlib', False)
            expected = '3.3.0'
            result = packages.installed_version('matplotlib')
            self.assertEqual(result, expected)

        # Test with non-existent package version
        with self.subTest():
            with self.assertRaises(SystemExit) as cm_:
                packages.install_package('pandas==100000')
            self.assertEqual(cm_.exception.code, 2)

        # Test package reinstall to more updated version
        with self.subTest():
            packages.install_package('matplotlib==3.3.1')
            expected = '3.3.1'
            result = packages.installed_version('matplotlib')
            self.assertEqual(result, expected)

        # Test reverting to outdated package version
        with self.subTest():
            packages.install_package('matplotlib==3.3.0', False)
            expected = '3.3.0'
            result = packages.installed_version('matplotlib')
            self.assertEqual(result, expected)

        # Test package reinstall to more updated version
        with self.subTest():
            packages.install_package('matplotlib==3.3.1')
            expected = '3.3.1'
            result = packages.installed_version('matplotlib')
            self.assertEqual(result, expected)

        # Test package reinstall to most recent version
        with self.subTest():
            packages.install_package('matplotlib>=3.3.1')
            expected = '3.3.1'
            result = packages.installed_version('matplotlib')
            self.assertGreater(result, expected)

    def test_install(self):
        """Unittest to test the install function."""
        # Test with undefined requirements directory
        with self.subTest():
            with self.assertRaises(SystemExit) as cm_:
                requirements_dir = data.hashstring(str(random()))
                packages.install(requirements_dir, self.venv_dir)
            self.assertEqual(cm_.exception.code, 3)

        # Test with default expected behaviour
        with self.subTest():
            # At least one expected package
            expected_package = 'Flask'
            expected = True

            # Create temporary directory
            packages.install(ROOT_DIR, self.venv_dir)

            # Get raw packages in requirements format
            pkgs = shared.run_script('python3 -m pip freeze')[1]

            # Get packages with versions removed
            installed_packages = [
                package.decode().split('==')[0] for package in pkgs.split()
                ]
            result = expected_package in installed_packages
            self.assertEqual(result, expected)

    def test_installed_version(self):
        """Unittest to test the installed_version function."""
        package = 'PattooShared'

        with self.subTest():
            # Uninstall package and verify
            shared.run_script(
                'python3 -m pip uninstall -y {}'.format(package))
            result = packages.installed_version(package)
            expected = None
            self.assertEqual(result, expected)

        with self.subTest():
            # Test with installed version
            shared.run_script(
                'python3 -m pip install {}==0.0.90'.format(package))
            result = packages.installed_version(package)
            expected = '0.0.90'
            self.assertEqual(result, expected)

    def test_package_details(self):
        """Unittest to test the package_details function."""
        # Initialise key variables
        Package = namedtuple('Package', 'name version inequality')
        packages_ = [
            'PattooShared>=0.0.106',
            'PyYAML',
            'Flask-Session',
            'PyMySQL',
            'SQLAlchemy',
            'graphene <= 2.7',
            'Flask',
            'Flask-CORS',
            'Flask-GraphQL',
            'Flask-GraphQL-Auth',
            'Flask-Caching',
            'Flask-Testing',
            'gunicorn',
            'requests',
            'Werkzeug',
            'numpy',
            'pandas',
            'psutil',
            'tblib',
            'distro',
            'python-gnupg==0.4.6',
        ]
        expected = [
            Package(name='PattooShared', version='0.0.106', inequality=True),
            Package(name='PyYAML', version=None, inequality=False),
            Package(name='Flask-Session', version=None, inequality=False),
            Package(name='PyMySQL', version=None, inequality=False),
            Package(name='SQLAlchemy', version=None, inequality=False),
            Package(name='graphene', version='2.7', inequality=True),
            Package(name='Flask', version=None, inequality=False),
            Package(name='Flask-CORS', version=None, inequality=False),
            Package(name='Flask-GraphQL', version=None, inequality=False),
            Package(name='Flask-GraphQL-Auth', version=None, inequality=False),
            Package(name='Flask-Caching', version=None, inequality=False),
            Package(name='Flask-Testing', version=None, inequality=False),
            Package(name='gunicorn', version=None, inequality=False),
            Package(name='requests', version=None, inequality=False),
            Package(name='Werkzeug', version=None, inequality=False),
            Package(name='numpy', version=None, inequality=False),
            Package(name='pandas', version=None, inequality=False),
            Package(name='psutil', version=None, inequality=False),
            Package(name='tblib', version=None, inequality=False),
            Package(name='distro', version=None, inequality=False),
            Package(name='python-gnupg', version='0.4.6', inequality=True),
        ]

        # Test
        for key, package in enumerate(packages_):
            result = packages.package_details(package)
            self.assertEqual(result, expected[key])


if __name__ == '__main__':
    # Make sure the environment is OK to run unittests
    UnittestConfig().create()

    # Do the unit test
    unittest.main()



"""Shared functions and methods utilized by the pattoo installation."""
# Standard imports
import sys
import os
import subprocess
import traceback
import shutil
import getpass


def run_script(cli_string, die=True, verbose=True):
    """Run the cli_string UNIX CLI command and record output.

    Args:
        cli_string: String of command to run
        die: Exit with error if True

    Returns:
        (returncode, stdoutdata, stderrdata):
            Execution code, STDOUT output and STDERR output.

    """
    # Initialize key variables
    messages = []
    stdoutdata = ''.encode()
    stderrdata = ''.encode()
    returncode = 1

    # Enable verbose mode if True
    if verbose is True:
        print('Running Command: "{}"'.format(cli_string))

    # Run update_targets script
    do_command_list = cli_string.split()

    # Create the subprocess object
    try:
        process = subprocess.Popen(
            do_command_list,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
        stdoutdata, stderrdata = process.communicate()
        returncode = process.returncode
    except:
        (exc_type, exc_value, exc_traceback) = sys.exc_info()
        messages.append('''\
Bug: Exception Type:{}, Exception Instance: {}, Stack Trace Object: {}]\
    '''.format(exc_type, exc_value, exc_traceback))
        messages.append(traceback.format_exc())

    # Crash if the return code is not 0
    if bool(returncode) is True:
        # Print the Return Code header
        messages.append(
            'Return code:{}'.format(returncode)
        )

        # Print the STDOUT
        for line in stdoutdata.decode().split('\n'):
            messages.append(
                'STDOUT: {}'.format(line)
            )

        # Print the STDERR
        for line in stderrdata.decode().split('\n'):
            messages.append(
                'STDERR: {}'.format(line)
            )

        # Log message
        if verbose is True:
            print('messages: {}'.format(messages))
        if bool(messages) is True:
            for log_message in messages:
                if verbose is True:
                    print(log_message)

            if bool(die) is True:
                # All done
                sys.exit(2)

    # Return
    return (returncode, stdoutdata, stderrdata)


def log(message):
    """Log messages and exit abnormally.

    Args:
        message: Message to print

    Returns:
        None

    """
    # exit
    print('\nPATTOO Error: {}'.format(message))
    sys.exit(3)


def chown(directory):
    """Recursively change the ownership of files in a directory.

    The directory must have the string '/pattoo/' in it

    Args:
        directory: Directory to create

    Returns:
        None

    """
    # Initialize key variables
    username = 'pattoo'
    group = 'pattoo'

    if getpass.getuser() != 'root':
        log('''\
Current user is not root, please execute script as root to continue''')

    # Change ownership
    if '{}pattoo'.format(os.sep) in directory:
        # Change the parent directory
        shutil.chown(directory, user=username, group=group)

        # Recursively change the sub-directories and files
        for root, dirs, files_ in os.walk(directory):
            for dir_ in dirs:
                shutil.chown(
                    os.path.join(root, dir_), user=username, group=group)
            for file_ in files_:
                shutil.chown(
                    os.path.join(root, file_), user=username, group=group)



"""Encryption module."""

import uuid
import hashlib
import os
import random
import string
import stat
from collections import namedtuple

# PIP imports
import gnupg

# Pattoo imports
from pattoo_shared import configuration
from pattoo_shared import log
from pattoo_shared import files

# Create constant for processing
_METADATA = namedtuple('_METADATA', 'fingerprint passphrase')


class KeyRing():
    """Class for managing PGP keyring."""

    def __init__(self, agent_name, directory=None):
        """Instantiate the class.

        Args:
            agent_name: Name of agent generating the private key
            directory: Alternative key storage directory

        Returns:
            None

        """
        # Initialize key variables
        config = configuration.Config()
        self._agent_name = agent_name

        # Use the agent_id as the email address because it is a unique
        # identifier across all agents. This allows multiple agents with the
        # same name to have indepenent sessions.
        self.email = files.get_agent_id(agent_name, config)

        # Associate directories
        if directory is None:
            # Store keys and keyrings in directories named after the agent.
            # If the same directory were universally used there could be
            # conflicts in cases where the API daemon and agents run on the
            # same server This eliminates this risk.
            keyring_directory = config.keyring_directory(agent_name)
            self._keys_directory = config.keys_directory(agent_name)
        else:
            keyring_directory = '{}{}.keyring'.format(directory, os.sep)
            self._keys_directory = directory
            files.mkdir(directory)
            files.mkdir(keyring_directory)

        # Initialize GPG object. Note: options=['--pinentry-mode=loopback']
        # ensures the passphrase can be entered via python and won't prompt
        # the user.
        self._gpg = gnupg.GPG(
            gnupghome=keyring_directory,
            options=['--pinentry-mode=loopback'])

        # Import metadata for managing keys
        metadata = self._import()

        # Create and store metadata if nonexistent
        if metadata is None:
            metadata = self._generate(self.email)
            self._export(metadata)

        # Share passphrase
        self._passphrase = metadata.passphrase
        self._fingerprint = metadata.fingerprint

        # Get the public key ID
        self._public_keyid = self._get_public_keyid(metadata.fingerprint)

    def _import(self):
        """Import private key information.

        Method to import the fingerprint and passphrase of the owned public /
        private key pair of the user. The method also looks for a wrapper on
        the file to distinguish the public private key pair the user currently
        owns.

        Args:
            None

        Returns:
            result: Tuple of (fingerprint, passphrase), None if nonexistent

        """
        # Initialize key variables
        result = None
        directory = self._keys_directory

        # Get list of key files whose names end with the wrapper
        key = [_file for _file in os.listdir(
            directory) if _file.endswith(self.email)]

        # Process data
        count = len(key)
        if count > 1:
            # Die if we have duplicate keys found in the directory
            log_message = '''\
More than one agent_id "{}" key-pair stores found'''.format(self.email)
            log.log2die(1062, log_message)

        elif count == 1:
            # Get passphrase
            filename = key[0]
            filepath = os.path.abspath(os.path.join(directory, filename))
            try:
                fh_ = open(filepath, 'r')
            except PermissionError:
                log.log2die(1091, '''\
Insufficient permissions for reading the file:{}'''.format(filepath))
            except:
                log.log2die(1093, 'Error reading file:{}'.format(filepath))
            else:
                with fh_:
                    passphrase = fh_.read()

            # Basic validation
            if isinstance(passphrase, str) is False:
                log.log2die(1094, 'Corrupted keyfile:{}'.format(filepath))

            # Retrieve fingerprint from the filename
            fingerprint = filename.split('-')[0]
            result = _METADATA(fingerprint=fingerprint, passphrase=passphrase)

        return result

    def _export(self, metadata):
        """Export pertinent information.

        Method to store the passphrase for future retrieval and name the file
        by the fingerprint of the class. The method also adds a wrapper to the
        name of the file to assist in searching for the public private key
        pair, it finds the pair it owns.

        Args:
            None

        Returns:
            None
        """
        # Initialize variables
        directory = self._keys_directory
        filepath = os.path.abspath(os.path.join(
            directory, '{0}-{1}'.format(metadata.fingerprint, self.email)))

        # Export data
        try:
            fh_ = open(filepath, 'w')
        except PermissionError:
            log.log2die(1095, '''\
Insufficient permissions for writing the file:{}'''.format(filepath))
        except:
            log.log2die(1096, 'Error writing file:{}'.format(filepath))
        else:
            with fh_:
                fh_.write(metadata.passphrase)

        # Allow RW file access to only the current user
        os.chmod(filepath, stat.S_IRUSR | stat.S_IWUSR)

    def _generate(self, email, key_type='RSA', key_length=4096):
        """Generate public / private key-pairs.

        Args:
            email (str): Email of the user
            key_type (str): The key type of the private public key pair
            key_length (int): Key length of the private public key pair

        Returns:
            None
        """
        # Initialize key variables
        comment = 'Key pair generated for {}'.format(self._agent_name)

        # Generates random passphrase
        passphrase = hashlib.sha256(str(uuid.uuid4()).encode()).hexdigest()

        # Generate key-pair
        data_ = self._gpg.gen_key_input(
            key_type=key_type,
            key_length=key_length,
            name_real=self._agent_name,
            name_comment=comment,
            name_email=email,
            passphrase=passphrase)
        key = self._gpg.gen_key(data_)
        fingerprint = key.fingerprint

        # Return results
        result = _METADATA(passphrase=passphrase, fingerprint=fingerprint)
        return result

    def _get_public_keyid(self, fingerprint):
        """Get the public key ID.

        Get the public key associated with the instance from a database of keys
        stored in the GnuPG keyring.

        Args:
            fingerprint: Fingerprint of the object instance

        Returns:
            result: Public key ID

        """
        # Initialize key variables
        result = None
        keys = self._gpg.list_keys()

        # Identify the public key
        for key in keys:
            if key['fingerprint'] == fingerprint:
                result = key['keyid']
                break

        # Return
        return result


class Encryption(KeyRing):
    """Class for managing PGP keypairs, currently using the python-gnupg.

    Resources:

    Chapter 7. Kurose J., Ross K.. (April, 2016).
    Computer Networking: A Top Down Approach.
    Pearson/Addison Wesley. Retrieved from
    http://www-net.cs.umass.edu/kurose-ross-ppt-7e/Chapter_8_V7.0.pptx

    Secure email style of encryption was used. So first, the key pairs
    are generated (public and private keys) so that it can be used to
    encrypt a randomly generated nonce and symmetric key. The nonce is
    used to provide a level of authentication. The symmetric key is
    used to encrypt the data since it is computationally less expensive
    than encrypting with a public key.

    What makes this module special, is it's ability to have a
    non-volatile approach to handle the passphrase, fingerprint
    and other variables with ease. The python-gnupg module
    provides most of the functionalities except keeping a track
    of the passphrase, the fingerprint, who owns the
    passphrase and fingerprint, etc. This module also simplifies
    the python-gnupg module to it can be widely used.

    The module uses one primary public private key pair to encrypt
    data. However, the module can also store other public keys.

    """

    def __init__(self, agent_name, directory=None):
        """Instantiate the class.

        Args:
            agent_name: Name of agent generating the private key
            email: Email address to use for encryption
            directory: Alternative key storage directory

        Returns:
            None

        """
        KeyRing.__init__(self, agent_name, directory=directory)

    def decrypt(self, data):
        """Decrypt data using a public key the GnuPG keyring.

        Args:
            data (str): Data in String ASCII to be decrypted

        Returns:
            result: Decrypted data into string

        """
        # Return
        result = self._decrypt(data)
        return result

    def sdecrypt(self, data, passphrase):
        """Decrypt data using an existing public key in the GnuPG keyring.

        Args:
            data (str): Data in ASCII string to be decrypted
            passphrase (str): Passphrase used in the encryption

        Returns:
            str: ASCII string of decrypted data
        """
        # Return
        result = self._decrypt(data, passphrase=passphrase)
        return result

    def _decrypt(self, data, passphrase=None):
        """Decrypt data using a public key the GnuPG keyring.

        Args:
            data (str): Data in String ASCII to be decrypted
            passphrase (str): Passphrase of the user

        Returns:
            result: Decrypted data into string

        """
        # Initialize the passphrase
        if bool(passphrase) is False:
            passphrase = self._passphrase
        decrypted = self._gpg.decrypt(data, passphrase=passphrase)

        # Return
        result = (decrypted.data).decode()
        return result

    def encrypt(self, data, recipients=None):
        """Encrypt data using expected key in the user's GnuPG keyring,

        Args:
            data (str): Data to be encrypted
            recipients (int): Fingerprint of recipient

        Returns:
            result (str): encrypted data in ASCII string

        """
        # Initialize key variables
        if recipients is None:
            recipients = self._fingerprint

        # Return
        result = str(
            self._gpg.encrypt(
                data,
                passphrase=None,
                recipients=recipients,
                symmetric=False
            )
        )
        return result

    def sencrypt(self, data, passphrase):
        """Symmetric encrypt.

        Method to encrypt data using symmmetric key encryption
        using a passphrase and encryption
        algorithm

        Args:
            data (str): String of data to be encrypted
            passphrase (str): Passphrase to be used to encrypt the data

        Returns:
            result: ASCII string of encrypted data

        """
        # Return
        result = str(
            self._gpg.encrypt(
                data,
                passphrase=passphrase,
                recipients=None,
                symmetric=True
            )
        )
        return result

    def trust(self, fingerprint, trustlevel='TRUST_ULTIMATE'):
        """Trust imported public key to encrypt data.

        Args:
            fingerprint (int): Fingerprint of pubic key to trust
            trustlevel (str): Trust level to assign to public key

        Returns:
            None
        """
        # Apply
        self._gpg.trust_keys(fingerprint, trustlevel)

    def fingerprint(self, email=None):
        """Get GnuPG keyring fingerprint that matches an email address.

        Args:
            email (str): Email address

        Returns:
            result: Fingerprint that is associated with the email

        """
        # Initialize key variables
        result = None

        # Go through each public key
        if email is None:
            result = self._fingerprint
        else:
            for key in self._gpg.list_keys():
                uids = list(filter((lambda item: email in item), key['uids']))
                if bool(uids) is True:
                    parts = uids[0].split()
                    wrapped_email = list(
                        filter((lambda item: '<' in item), parts))
                    unwrapped_email = wrapped_email[0].strip('<>')
                    if unwrapped_email == email:
                        result = key['fingerprint']
                        break

        return result

    def pdelete(self, fingerprint):
        """Delete public key from keyring.

        Args:
            fingerprint: Fingerprint of public key to be deleted

        Returns:
            result (bool): True if the public key was deleted False if the
                public key was not deleted

        """
        # Initialize key variables
        result = False

        # Delete public key
        if fingerprint != self._fingerprint:
            _result = self._gpg.delete_keys(fingerprint)
            result = str(_result).lower() == 'ok'
        return result

    def pexport(self, recipient=None):
        """Export the user's public key into ASCII.

        Args:
            recipient: fingerpint for which public key is required

        Returns:
            result: String of ASCII armored public key

        """
        # Initialize key variables
        result = None

        # Return result
        if recipient is None:
            keyid = self._public_keyid
        else:
            keyid = self._get_public_keyid(recipient)
        if keyid is not None:
            result = self._gpg.export_keys(keyid)
        return result

    def pimport(self, key):
        """Import public key into the current user's GnuPG keyring.

        Args:
            key (str): String of public key armored ASCII

        Returns:
            None

        """
        # Import
        self._gpg.import_keys(key)


def generate_key(length=70):
    """Generate a random string for use as a encryption key.

    Args:
        length (str): Maximum length of the key.

    Returns:
        result: key

    """
    # Return
    characters = string.ascii_letters + string.digits + string.punctuation
    result = ''.join(random.choice(characters) for i in range(length))
    return result



#!/usr/bin/env/python3
"""Test pattoo configuration script."""

import os
import unittest
import sys
import tempfile
from random import random

import yaml
import distro

# Try to create a working PYTHONPATH
EXEC_DIR = os.path.dirname(os.path.realpath(__file__))
ROOT_DIR = os.path.abspath(os.path.join(
    os.path.abspath(os.path.join(
        os.path.abspath(os.path.join(
            EXEC_DIR, os.pardir)), os.pardir)), os.pardir))
_EXPECTED = '''\
{0}pattoo-shared{0}tests{0}pattoo_shared_{0}installation'''.format(os.sep)
if EXEC_DIR.endswith(_EXPECTED) is True:
    # We need to prepend the path in case PattooShared has been installed
    # elsewhere on the system using PIP. This could corrupt expected results
    sys.path.insert(0, ROOT_DIR)
else:
    print('''This script is not installed in the "{0}" directory. Please fix.\
'''.format(_EXPECTED))
    sys.exit(2)

# Pattoo imports
from tests.libraries.configuration import UnittestConfig
from pattoo_shared import data
from pattoo_shared.installation import shared
from pattoo_shared.installation.systemd import _filepaths, remove_file
from pattoo_shared.installation.systemd import _get_runtime_directory
from pattoo_shared.installation.systemd import symlink_dir, _check_symlinks


class Test_Systemd(unittest.TestCase):
    """Checks all functions and methods."""

    def test__filepaths(self):
        """Testing method or function named "_filepaths"."""
        # Initialize temporary directory
        with tempfile.TemporaryDirectory() as temp_dir:
            file_path = os.path.join(temp_dir, 'test_file.txt')

            # Create file
            with open(file_path, 'w+') as test_file:
                test_file.write('this is a test')
            expected = [file_path]
            result = _filepaths(temp_dir)
            self.assertEqual(expected, result)

    def test_symlink_dir(self):
        """Testing method or function named "symlink_dir"."""
        # Initialise key variables
        linux_distro = distro.linux_distribution()[0].lower()
        etc_dir = '/etc/systemd/system/multi-user.target.wants'

        if linux_distro in ['ubuntu']:
            expected = '/lib/systemd/system'
        else:
            # Expected directory for CentOS
            expected = '/usr/lib/systemd/system'

        # Test directory without symlinks
        with tempfile.TemporaryDirectory() as temp_dir:
            with self.assertRaises(SystemExit) as cm_:
                symlink_dir(temp_dir)
            self.assertEqual(cm_.exception.code, 3)

        # Test directory with expected symlinks
        result = symlink_dir(etc_dir)
        self.assertEqual(result, expected)

    def test__get_runtime_directory_default(self):
        """Testing method or function named "_get_runtime_directory"."""
        # Initialize key variables
        default = {
            'pattoo': {
                'language': 'en',
                'log_directory': '/var/log/pattoo',
                'log_level': 'debug',
                'cache_directory': '/opt/pattoo-cache',
                'daemon_directory': '/opt/pattoo-daemon',
                'system_daemon_directory': '/var/run/pattoo'
            },
            'pattoo_agent_api': {
                'ip_address': '127.0.0.1',
                'ip_bind_port': 20201
            },
            'pattoo_web_api': {
                'ip_address': '127.0.0.1',
                'ip_bind_port': 20202,
            }
        }

        expected = ('/var/run/pattoo', 'pattoo')

        # Retrieve runtime directory from temp directory
        with tempfile.TemporaryDirectory() as temp_dir:

            # Test with default system daemon directory
            with open(os.path.join(temp_dir, 'pattoo.yaml'), 'w+') as fh_:
                yaml.safe_dump(default, stream=fh_, default_flow_style=False)
                result = _get_runtime_directory(temp_dir)
            self.assertEqual(expected, result)

    def test__get_runtime_undefined_file(self):
        """Testing method or function named "_get_runtime_directory"."""
        with self.assertRaises(SystemExit) as cm_:
            # Generate random directory name
            _get_runtime_directory(data.hashstring(str(random())))
        self.assertEqual(cm_.exception.code, 3)

    def test__get_runtime_directory_no_systemd(self):
        """Testing method or function named "_get_runtime_directory"."""
        # Initialize key variables
        fake = {
            'pattoo': {
                'language': 'en',
                'log_directory': '/var/log/pattoo',
                'log_level': 'debug',
                'cache_directory': '/opt/pattoo-cache',
                'daemon_directory': '/opt/pattoo-daemon'
            }
        }

        # Retrieve runtime directory from temp directory
        with tempfile.TemporaryDirectory() as temp_dir:
            with open(os.path.join(temp_dir, 'pattoo.yaml'), 'w+') as fh_:
                yaml.safe_dump(fake, stream=fh_, default_flow_style=False)

            # Test without system daemon directory
            with self.assertRaises(SystemExit) as cm_:
                _get_runtime_directory(temp_dir)
            self.assertEqual(cm_.exception.code, 3)

    def test___check_symlinks(self):
        """Testing method or function named "_get_runtime_directory"."""
        # Initialize key variables
        daemons = [
            'pattoo_apid',
            'pattoo_api_agentd',
            'pattoo_ingesterd'
        ]

        with tempfile.TemporaryDirectory() as temp_dir:
            # Create target directory
            target_dir = os.path.join(temp_dir, 'test_symlink')
            os.mkdir(target_dir)
            # Create service files
            for daemon in daemons:
                daemon_path = os.path.join(
                    target_dir, '{}.service'.format(daemon))
                with open(daemon_path, 'w'):
                    pass

                # Since this block will exit if the user is not running as root
                # A try except was used to test if the files are deleted
                with self.subTest():
                    try:
                        _check_symlinks(target_dir, daemon)
                    except SystemExit:

                        # If the files are successfully deleted, all elements
                        # of this list will be false
                        result = [
                            os.path.exists(daemon_path) for daemon in daemons]
                        self.assertTrue(not any(result))
                # Check for symlinks and sudo access
                with self.assertRaises(SystemExit) as cm_:
                    _check_symlinks(temp_dir, daemon)
                self.assertEqual(cm_.exception.code, 2)

    def test_remove_file(self):
        """Testing method named "remove_file"."""
        with tempfile.TemporaryDirectory() as temp_dir:
            file_path = os.path.join(temp_dir, 'test.txt')

            # Testing function when the file does not exist
            with self.subTest():
                with self.assertRaises(SystemExit) as cm_:
                    remove_file(file_path)
                self.assertEqual(cm_.exception.code, 2)

            # Testing function by creating file and then removing it
            with self.subTest():
                with open(file_path, 'w'):
                    pass
                remove_file(file_path)
                self.assertFalse(os.path.exists(file_path))


if __name__ == '__main__':
    # Make sure the environment is OK to run unittests
    UnittestConfig().create()

    # Do the unit test
    unittest.main()



#!/usr/bin/env python3
"""Pattoo classes that manage various configurations."""

# Standard imports
import os
import stat

# Import project libraries
from pattoo_shared import files
from pattoo_shared import log
from pattoo_shared import url
from pattoo_shared.constants import (
    PATTOO_API_AGENT_PREFIX)
from pattoo_shared.variables import PollingPoint


def _config_reader(filename):
    """Read a configuration file.

    Args:
        filename: Name of file to read

    Returns:
        config_dict: Dict representation of YAML in the file

    """
    # Get the configuration directory
    # Expand linux ~ notation for home directories if provided.
    _config_directory = log.check_environment()
    config_directory = os.path.expanduser(_config_directory)
    config_file = '{}{}{}'.format(config_directory, os.sep, filename)
    config_dict = files.read_yaml_file(config_file)
    return config_dict


class BaseConfig():
    """Class gathers all configuration information."""

    def __init__(self):
        """Initialize the class.

        Args:
            None

        Returns:
            None

        """
        # Read data
        self._base_yaml_configuration = _config_reader('pattoo.yaml')

    def config_directory(self):
        """Get config_directory.

        Args:
            None

        Returns:
            result: result

        """
        # Return
        result = log.check_environment()
        return result

    def log_directory(self):
        """Get log_directory.

        Args:
            None

        Returns:
            result: result

        """
        # Get result
        sub_key = 'log_directory'
        result = None
        key = 'pattoo'

        # Get new result
        _result = search(key, sub_key, self._base_yaml_configuration)

        # Expand linux ~ notation for home directories if provided.
        result = os.path.expanduser(_result)

        # Check if value exists. We cannot use log2die_safe as it does not
        # require a log directory location to work properly
        if os.path.isdir(result) is False:
            log_message = (
                'log_directory: "{}" '
                'in configuration doesn\'t exist!'.format(result))
            log.log2die_safe(1003, log_message)

        # Return
        return result

    def log_file(self):
        """Get log_file.

        Args:
            None

        Returns:
            result: result

        """
        _log_directory = self.log_directory()
        result = '{}{}pattoo.log'.format(_log_directory, os.sep)
        return result

    def log_file_api(self):
        """Get log_file_api.

        Args:
            None

        Returns:
            result: result

        """
        # Get result
        _log_directory = self.log_directory()
        result = '{}{}pattoo-api.log'.format(_log_directory, os.sep)
        return result

    def log_file_daemon(self):
        """Get log_file_daemon.

        Args:
            None

        Returns:
            result: result

        """
        # Get result
        _log_directory = self.log_directory()
        result = '{}{}pattoo-daemon.log'.format(_log_directory, os.sep)
        return result

    def log_level(self):
        """Get log_level.

        Args:
            None

        Returns:
            result: result

        """
        # Get result
        sub_key = 'log_level'
        key = 'pattoo'
        result = None

        # Return
        intermediate = search(
            key, sub_key, self._base_yaml_configuration, die=False)
        if intermediate is None:
            result = 'debug'
        else:
            result = '{}'.format(intermediate).lower()
        return result

    def cache_directory(self):
        """Determine the cache_directory.

        Args:
            None

        Returns:
            value: configured cache_directory

        """
        # Initialize key variables
        key = 'pattoo'
        sub_key = 'cache_directory'

        # Get result
        _value = search(key, sub_key, self._base_yaml_configuration)

        # Expand linux ~ notation for home directories if provided.
        value = os.path.expanduser(_value)

        # Create directory if it doesn't exist
        files.mkdir(value)

        # Return
        return value

    def agent_cache_directory(self, agent_program):
        """Get agent_cache_directory.

        Args:
            agent_program: Name of agent

        Returns:
            result: result

        """
        # Get result
        result = '{}/{}'.format(self.cache_directory(), agent_program)

        # Create directory if it doesn't exist
        files.mkdir(result)

        # Return
        return result

    def daemon_directory(self):
        """Determine the daemon_directory.

        Args:
            None

        Returns:
            value: configured daemon_directory

        """
        # Initialize key variables
        key = 'pattoo'
        sub_key = 'daemon_directory'

        # Get result
        _value = search(key, sub_key, self._base_yaml_configuration)

        # Expand linux ~ notation for home directories if provided.
        value = os.path.expanduser(_value)

        # Create directory if it doesn't exist
        files.mkdir(value)

        # Return
        return value

    def keyring_directory(self, agent_name):
        """Get keyring_directory.

        Args:
            agent_name: Name of agent

        Returns:
            result: result

        """
        # Get result
        result = '{0}{1}.keyring'.format(
            self.keys_directory(agent_name), os.sep)

        # Create directory if it doesn't exist
        files.mkdir(result)

        # Make only accessible to the user
        os.chmod(result, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)

        # Return
        return result

    def keys_directory(self, agent_name):
        """Get keys_directory.

        Args:
            agent_name: Name of agent

        Returns:
            result: result

        """
        # Get result
        result = '{0}{1}keys{1}{2}'.format(
            self.daemon_directory(), os.sep, agent_name)

        # Create directory if it doesn't exist
        files.mkdir(result)

        # Make only accessible to the user
        os.chmod(result, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)

        # Return
        return result

    def system_daemon_directory(self):
        """Determine the system_daemon_directory.

        Args:
            None

        Returns:
            value: configured system_daemon_directory

        """
        # Initialize key variables
        key = 'pattoo'
        sub_key = 'system_daemon_directory'

        # Get result
        result = search(key, sub_key, self._base_yaml_configuration, die=False)
        _value = result if bool(result) else self.daemon_directory()

        # Expand linux ~ notation for home directories if provided.
        value = os.path.expanduser(_value)

        # Create directory if it doesn't exist
        files.mkdir(value)

        # Return
        return value

    def language(self):
        """Get language.

        Args:
            None

        Returns:
            result: result

        """
        # Get result
        key = 'pattoo'
        sub_key = 'language'
        intermediate = search(
            key, sub_key, self._base_yaml_configuration, die=False)

        # Default to 'en'
        if bool(intermediate) is False:
            result = 'en'
        else:
            result = str(intermediate).lower()
        return result


class ServerConfig(BaseConfig):
    """Class gathers all configuration information."""

    def __init__(self):
        """Initialize the class.

        Args:
            None

        Returns:
            None

        """
        # Get the configuration
        BaseConfig.__init__(self)
        self._server_yaml_configuration = _config_reader('pattoo_server.yaml')


class Config(BaseConfig):
    """Class gathers all configuration information."""

    def __init__(self):
        """Initialize the class.

        Args:
            None

        Returns:
            None

        """
        # Get the configuration
        BaseConfig.__init__(self)

        self._agent_yaml_configuration = _config_reader('pattoo_agent.yaml')

    def agent_api_ip_address(self):
        """Get api_ip_address.

        Args:
            None

        Returns:
            result: result

        """
        # Initialize key variables
        key = 'pattoo_agent_api'
        sub_key = 'ip_address'

        # Get result
        result = search(
            key, sub_key, self._agent_yaml_configuration, die=False)
        if result is None:
            result = 'localhost'
        return result

    def agent_api_ip_bind_port(self):
        """Get agent_api_ip_bind_port.

        Args:
            None

        Returns:
            result: result

        """
        # Initialize key variables
        key = 'pattoo_agent_api'
        sub_key = 'ip_bind_port'

        # Get result
        intermediate = search(
            key, sub_key, self._agent_yaml_configuration, die=False)
        if intermediate is None:
            result = 20201
        else:
            result = int(intermediate)
        return result

    def agent_api_uri(self):
        """Get agent_api_uri.

        Args:
            None

        Returns:
            result: result

        """
        # Return
        result = '{}/receive'.format(PATTOO_API_AGENT_PREFIX)
        return result

    def agent_api_key(self):
        """Get URL for key exchange.

        Args:
            None

        Returns:
            url (str): URL of the key exchange point
        """
        url_ = '{}/key'.format(PATTOO_API_AGENT_PREFIX)
        return url_

    def agent_api_validation(self):
        """Get URL to validate encryption status.

        Args:
            None

        Returns:
            url (str): URL of the validation point
        """
        url_ = '{}/validation'.format(PATTOO_API_AGENT_PREFIX)
        return url_

    def agent_api_encrypted(self):
        """Get URL to receive encrypted data.

        Args:
            None

        Returns:
            result: result

        """
        # Return
        result = '{}/encrypted'.format(PATTOO_API_AGENT_PREFIX)
        return result

    def agent_api_server_url(self, agent_id):
        """Get pattoo server's remote URL.

        Args:
            agent_id: Agent ID

        Returns:
            result: URL.

        """
        # Return
        _ip = url.url_ip_address(self.agent_api_ip_address())
        result = (
            'http://{}:{}{}/{}'.format(
                _ip,
                self.agent_api_ip_bind_port(),
                self.agent_api_uri(), agent_id))
        return result

    def agent_api_key_url(self):
        """Exchange point for public keys.

        Args:
            None

        Returns:
            link (str): Link of the key exchange point

        """
        # Initialize key variables
        _ip = url.url_ip_address(self.agent_api_ip_address())
        link = (
            'http://{}:{}{}'.format(
                _ip,
                self.agent_api_ip_bind_port(),
                self.agent_api_key()
                )
            )

        return link

    def agent_api_validation_url(self):
        """Validation point for encryption.

        Args:
            None

        Returns:
            link (str): Link of the validation point

        """

        _ip = url.url_ip_address(self.agent_api_ip_address())
        link = (
            'http://{}:{}{}'.format(
                _ip,
                self.agent_api_ip_bind_port(),
                self.agent_api_validation()
                )
            )

        return link

    def agent_api_encrypted_url(self):
        """Encrypted data reception point.

        Args:
            None

        Returns:
            link (str): Link of encrypted data receive point

        """

        _ip = url.url_ip_address(self.agent_api_ip_address())
        link = (
            'http://{}:{}{}'.format(
                _ip,
                self.agent_api_ip_bind_port(),
                self.agent_api_encrypted()
                )
            )

        return link


def agent_config_filename(agent_program):
    """Get the configuration file name.

    Args:
        agent_program: Agent program name

    Returns:
        result: Name of file

    """
    # Get the configuration directory
    # Expand linux ~ notation for home directories if provided.
    _config_directory = log.check_environment()
    config_directory = os.path.expanduser(_config_directory)
    result = '{}{}{}.yaml'.format(config_directory, os.sep, agent_program)
    return result


def get_polling_points(_data):
    """Create list of PollingPoint objects.

    Args:
        _data: List of dicts with 'address' and 'multiplier' as keys

    Returns:
        results: List of PollingPoint objects

    """
    # Start conversion
    results = []

    if isinstance(_data, list) is True:
        # Cycle through list
        for item in _data:
            # Reject non dict data
            if isinstance(item, dict) is False:
                continue

            # Assign address value only if present
            if 'address' in item:
                address = item['address']
            else:
                continue

            # Assign replacement multiplier
            multiplier = item.get('multiplier', 1)

            # Populate result
            result = PollingPoint(address=address, multiplier=multiplier)
            results.append(result)

    # Return
    return results


def search(key, sub_key, config_dict, die=True):
    """Get config parameter from YAML.

    Args:
        key: Primary key
        sub_key: Secondary key
        config_dict: Dictionary to explore
        die: Die if true and the result encountered is None

    Returns:
        result: result

    """
    # Get result
    result = None

    # Verify config_dict is indeed a dict.
    # Die safely as log_directory is not defined
    if isinstance(config_dict, dict) is False:
        log.log2die_safe(1021, 'Invalid configuration file. YAML not found')

    # Get new result
    if config_dict.get(key) is not None:
        # Make sure we don't have a None value
        if config_dict[key] is None:
            log_message = (
                '{}: value in configuration is blank. Please fix'.format(key))
            log.log2die_safe(1004, log_message)

        # Get value we need
        result = config_dict[key].get(sub_key)

    # Error if not configured
    if result is None and die is True:
        log_message = (
            '{}:{} not defined in configuration dict {}'.format(
                key, sub_key, config_dict))
        log.log2die_safe(1016, log_message)

    # Return
    return result



#!/usr/bin/env python3
"""Test the files module."""

# Standard imports
import unittest
import os
import sys
import socket
import time

# Try to create a working PYTHONPATH
EXEC_DIR = os.path.dirname(os.path.realpath(__file__))
ROOT_DIR = os.path.abspath(os.path.join(
    os.path.abspath(os.path.join(EXEC_DIR, os.pardir)), os.pardir))
_EXPECTED = '{0}pattoo-shared{0}tests{0}pattoo_shared_'.format(os.sep)
if EXEC_DIR.endswith(_EXPECTED) is True:
    # We need to prepend the path in case PattooShared has been installed
    # elsewhere on the system using PIP. This could corrupt expected results
    sys.path.insert(0, ROOT_DIR)
else:
    print('''This script is not installed in the "{0}" directory. Please fix.\
'''.format(_EXPECTED))
    sys.exit(2)

# Pattoo imports
from pattoo_shared import variables
from pattoo_shared import files
from pattoo_shared.configuration import Config
from pattoo_shared.constants import (
    DATA_INT, DATA_STRING, DATA_FLOAT, DATAPOINT_KEYS, AGENT_METADATA_KEYS)
from pattoo_shared.variables import (
    DataPoint, DataPointMetadata, ConverterMetadata, PostingDataPoints,
    TargetDataPoints, TargetPollingPoints,
    PollingPoint, IPTargetPollingPoints, AgentPolledData, AgentAPIVariable)
from tests.libraries.configuration import UnittestConfig


class TestPostingDataPoints(unittest.TestCase):
    """Checks all functions and methods."""

    #########################################################################
    # General object setup
    #########################################################################

    def test___init__(self):
        """Testing function __init__."""
        # Setup PostingDataPoints - Valid
        agent_id = '1234'
        polling_interval = 10
        _datapoints = {
            'datapoint_pairs': [[0], [1]],
            'key_value_pairs': {
                '0': ['key_01', 'value_01'],
                '1': ['key_02', 'value_02']}
            }
        result = PostingDataPoints(agent_id, polling_interval, _datapoints)
        self.assertEqual(result.pattoo_agent_id, agent_id)
        self.assertEqual(result.pattoo_datapoints, _datapoints)
        self.assertEqual(
            result.pattoo_agent_polling_interval, polling_interval)
        self.assertTrue(result.valid)

        # Setup PostingDataPoints - Invalid
        for agent_id in [1, True, None, False, {1: 2}, [1, 2]]:
            polling_interval = 10
            result = PostingDataPoints(agent_id, polling_interval, _datapoints)
            self.assertEqual(result.pattoo_agent_id, agent_id)
            self.assertEqual(result.pattoo_datapoints, _datapoints)
            self.assertEqual(
                result.pattoo_agent_polling_interval, polling_interval)
            self.assertFalse(result.valid)

        for polling_interval in ['1', True, None, False, {1: 2}, [1, 2]]:
            agent_id = '1234'
            result = PostingDataPoints(agent_id, polling_interval, _datapoints)
            self.assertEqual(result.pattoo_agent_id, agent_id)
            self.assertEqual(result.pattoo_datapoints, _datapoints)
            self.assertEqual(
                result.pattoo_agent_polling_interval, polling_interval)
            self.assertFalse(result.valid)

        for datapoints in [1, True, None, False, {1: 2}, [1, 2]]:
            agent_id = '1234'
            polling_interval = 10
            result = PostingDataPoints(agent_id, polling_interval, datapoints)
            self.assertEqual(result.pattoo_agent_id, agent_id)
            self.assertEqual(result.pattoo_datapoints, datapoints)
            self.assertEqual(
                result.pattoo_agent_polling_interval, polling_interval)
            self.assertFalse(result.valid)


class TestDataPointMetadata(unittest.TestCase):
    """Checks all functions and methods."""

    #########################################################################
    # General object setup
    #########################################################################

    def test___init__(self):
        """Testing function __init__."""
        # Setup DataPoint - Valid
        for key, value in [
                (1, 2), ('1', 2), (1, '2'), ('1', '2'),
                (1.1, 2.1), ('1.1', 2.1), (1.1, '2.1'), ('1.1', '2.1')]:
            result = DataPointMetadata(key, value)
            self.assertEqual(result.key, str(key))
            self.assertEqual(result.value, str(value))
            self.assertTrue(result.valid)

        # Setup DataPoint - Invalid
        for key, value in [
                ('pattoo', 1), ('123pattoo123', 1),
                (None, 2), ('1', None), (True, '2'), ('1', True),
                ({}, 2.1), ('1.1', {2: 1}), (False, '2.1'), ('1.1', False)]:
            result = DataPointMetadata(key, value)
            self.assertFalse(result.valid)

    def test___repr__(self):
        """Testing function __repr__."""
        # Setup DataPointMetadata
        variable = DataPointMetadata(5, 6)

        # Test
        expected = ('''<DataPointMetadata key='5', value='6'>''')
        result = variable.__repr__()
        self.assertEqual(result, expected)


class TestConverterMetadata(unittest.TestCase):
    """Checks all functions and methods."""

    #########################################################################
    # General object setup
    #########################################################################

    def test___init__(self):
        """Testing function __init__."""
        # Setup DataPoint - Valid
        value = 1
        for key in AGENT_METADATA_KEYS:
            result = ConverterMetadata(key, value)
            self.assertEqual(result.key, str(key))
            self.assertEqual(result.value, str(value))
            self.assertTrue(result.valid)

        # Setup DataPoint - Invalid
        for key, value in [
                ('pattoo', 1), ('123pattoo123', 1),
                (None, 2), ('1', None), (True, '2'), ('1', True),
                ({}, 2.1), ('1.1', {2: 1}), (False, '2.1'), ('1.1', False)]:
            result = ConverterMetadata(key, value)
            self.assertFalse(result.valid)

    def test___repr__(self):
        """Testing function __repr__."""
        # Setup ConverterMetadata
        key = AGENT_METADATA_KEYS[0]
        variable = ConverterMetadata(key, 6)

        # Test
        expected = ('''<ConverterMetadata key='{}', value='6'>'''.format(key))
        result = variable.__repr__()
        self.assertEqual(result, expected)


class TestDataPoint(unittest.TestCase):
    """Checks all functions and methods."""

    #########################################################################
    # General object setup
    #########################################################################

    def test___init__(self):
        """Testing function __init__."""
        # Setup DataPoint - Valid
        value = 1093454
        _key_ = 'testing'
        _metakey = '_{}'.format(_key_)
        timestamp = int(time.time() * 1000)
        data_type = DATA_INT
        variable = DataPoint(_key_, value, data_type=data_type)
        variable.add(DataPointMetadata(_metakey, _metakey))

        # Test each variable
        self.assertEqual(variable.data_type, data_type)
        self.assertEqual(variable.value, value)
        self.assertEqual(variable.key, _key_)
        self.assertTrue(variable.timestamp >= timestamp)
        self.assertEqual(len(variable.checksum), 64)
        self.assertEqual(variable.checksum, '''\
306353a04200e3b889b18c6f78dd8e56a63a287218ec8424e22d31b4b961a905''')
        self.assertEqual(variable.valid, True)

        # Setup DataPoint - Valid
        value = 1093454
        timestamp = 7
        _key_ = 'testing'
        _metakey = '_{}'.format(_key_)
        data_type = DATA_INT
        variable = DataPoint(
            _key_, value, data_type=data_type, timestamp=timestamp)
        variable.add(DataPointMetadata(_metakey, _metakey))

        # Test each variable
        self.assertEqual(variable.data_type, data_type)
        self.assertEqual(variable.value, value)
        self.assertEqual(variable.key, _key_)
        self.assertEqual(variable.timestamp, timestamp)
        self.assertEqual(len(variable.checksum), 64)
        self.assertEqual(variable.checksum, '''\
306353a04200e3b889b18c6f78dd8e56a63a287218ec8424e22d31b4b961a905''')
        self.assertEqual(variable.valid, True)

        # Add metadata that should be ignored.
        for key in DATAPOINT_KEYS:
            variable.add(DataPointMetadata(key, '_{}_'.format(key)))
        variable.add(DataPointMetadata(_metakey, _metakey))

        # Test each variable (unchanged)
        self.assertEqual(variable.data_type, data_type)
        self.assertEqual(variable.value, value)
        self.assertEqual(variable.key, _key_)
        self.assertEqual(len(variable.checksum), 64)
        self.assertEqual(variable.checksum, '''\
306353a04200e3b889b18c6f78dd8e56a63a287218ec8424e22d31b4b961a905''')
        self.assertEqual(variable.valid, True)

        # Setup DataPoint - invalid data_type
        value = 1093454
        _key_ = 'testing'
        data_type = 123
        variable = DataPoint(_key_, value, data_type=data_type)

        # Test each variable
        self.assertEqual(variable.data_type, data_type)
        self.assertEqual(variable.value, value)
        self.assertEqual(variable.key, _key_)
        self.assertEqual(variable.valid, False)

        # Setup DataPoint - invalid value for numeric data_type
        value = '_123'
        _key_ = 'testing'
        data_type = DATA_INT
        variable = DataPoint(_key_, value, data_type=data_type)

        # Test each variable
        self.assertEqual(variable.data_type, data_type)
        self.assertEqual(variable.value, value)
        self.assertEqual(variable.key, _key_)
        self.assertEqual(variable.valid, False)

        # Setup DataPoint - valid value for integer data_type but
        # string for value
        value = '1093454'
        _key_ = 'testing'
        data_type = DATA_INT
        variable = DataPoint(_key_, value, data_type=data_type)

        # Test each variable
        self.assertEqual(variable.data_type, data_type)
        self.assertEqual(variable.value, int(value))
        self.assertEqual(variable.key, _key_)
        self.assertEqual(len(variable.checksum), 64)
        self.assertEqual(variable.checksum, '''\
7f99301d9be275b14af5626ffabe22a154415ed2ef7dad37f1707bd25b6afdc6''')
        self.assertEqual(variable.valid, True)

        # Setup DataPoint - valid value for int data_type but
        # string for value
        value = '1093454.3'
        _key_ = 'testing'
        data_type = DATA_INT
        variable = DataPoint(_key_, value, data_type=data_type)

        # Test each variable
        self.assertEqual(variable.data_type, data_type)
        self.assertEqual(variable.value, int(float(value)))
        self.assertEqual(variable.key, _key_)
        self.assertEqual(len(variable.checksum), 64)
        self.assertEqual(variable.checksum, '''\
7f99301d9be275b14af5626ffabe22a154415ed2ef7dad37f1707bd25b6afdc6''')
        self.assertEqual(variable.valid, True)

        # Setup DataPoint - valid value for int data_type but
        # string for value
        value = '1093454.3'
        _key_ = 'testing'
        data_type = DATA_FLOAT
        variable = DataPoint(_key_, value, data_type=data_type)

        # Test each variable
        self.assertEqual(variable.data_type, data_type)
        self.assertEqual(variable.value, float(value))
        self.assertEqual(variable.key, _key_)
        self.assertEqual(len(variable.checksum), 64)
        self.assertEqual(variable.checksum, '''\
ab48bdc902e2ea5476a54680a7ace0971ab90edb3f6ffe00a89b2d1e17b1548d''')
        self.assertEqual(variable.valid, True)

        # Setup DataPoint - valid value for str data_type
        for value in [0, 1, '1093454.3']:
            _key_ = 'testing'
            data_type = DATA_STRING
            variable = DataPoint(_key_, value, data_type=data_type)

            # Test each variable
            self.assertEqual(variable.data_type, data_type)
            self.assertEqual(variable.value, str(value))
            self.assertEqual(variable.key, _key_)
            self.assertEqual(len(variable.checksum), 64)
            self.assertEqual(variable.checksum, '''\
431111472993bf4d9b8b347476b79321fea8a337f3c1cb2fedaa185b54185540''')
            self.assertEqual(variable.valid, True)

        # Setup DataPoint - invalid value for str data_type
        for value in [True, False, None]:
            _key_ = 'testing'
            data_type = DATA_STRING
            variable = DataPoint(_key_, value, data_type=data_type)

            # Test each variable
            self.assertEqual(variable.data_type, data_type)
            self.assertEqual(variable.valid, False)
            self.assertEqual(variable.value, str(value))
            self.assertIsNone(variable.key)
            self.assertEqual(len(variable.checksum), 64)
            self.assertEqual(variable.checksum, '''\
a783370f88d8c54b5f5e6641af69d86dae5d4d62621d55cf7e63f6c66644c214''')

    def test___repr__(self):
        """Testing function __repr__."""
        # Need to see all the string output
        self.maxDiff = None

        # Setup DataPoint
        value = 10
        _key_ = 'testing'
        data_type = DATA_INT
        variable = DataPoint(_key_, value, data_type=data_type)

        # Test
        expected = ('''\
<DataPoint key='testing', value=10, data_type=99, \
timestamp={}, valid=True>'''.format(variable.timestamp))
        result = variable.__repr__()
        self.assertEqual(result, expected)

    def test_add(self):
        """Testing function add."""
        # Setup DataPoint - Valid
        value = 1093454
        _key_ = 'testing'
        data_type = DATA_INT
        variable = DataPoint(_key_, value, data_type=data_type)

        # Test adding
        for key, value in [(1, 2), (3, 4), (5, 6)]:
            metadata = DataPointMetadata(key, value)
            variable.add(metadata)
            self.assertEqual(variable.metadata[str(key)], str(value))

        self.assertEqual(len(variable.metadata), 3)
        self.assertEqual(len(variable.checksum), 64)
        self.assertEqual(variable.checksum, '''\
73ce7225ca1ea55f53c96991c9922a185cf695224b94f2051b8a853049ba1935''')

        # Test adding duplicates (no change)
        for key, value in [(1, 2), (3, 4), (5, 6)]:
            metadata = DataPointMetadata(key, value)
            variable.add(metadata)
            self.assertEqual(variable.metadata[str(key)], str(value))

        self.assertEqual(len(variable.metadata), 3)
        self.assertEqual(len(variable.checksum), 64)
        self.assertEqual(variable.checksum, '''\
73ce7225ca1ea55f53c96991c9922a185cf695224b94f2051b8a853049ba1935''')

        # Test adding with now update_checksum set to False. No change
        for key, value in [(10, 20), (30, 40), (50, 60)]:
            metadata = DataPointMetadata(key, value, update_checksum=False)
            variable.add(metadata)
            self.assertEqual(variable.metadata[str(key)], str(value))

        self.assertEqual(len(variable.metadata), 6)
        self.assertEqual(len(variable.checksum), 64)
        self.assertEqual(variable.checksum, '''\
73ce7225ca1ea55f53c96991c9922a185cf695224b94f2051b8a853049ba1935''')

        # Test adding with now update_checksum set to True. No change,
        # as they have been added already.
        for key, value in [(10, 20), (30, 40), (50, 60)]:
            metadata = DataPointMetadata(key, value, update_checksum=True)
            variable.add(metadata)
            self.assertEqual(variable.metadata[str(key)], str(value))

        self.assertEqual(len(variable.metadata), 6)
        self.assertEqual(len(variable.checksum), 64)
        self.assertEqual(variable.checksum, '''\
73ce7225ca1ea55f53c96991c9922a185cf695224b94f2051b8a853049ba1935''')

        # Test adding with now update_checksum set to True
        for key, value in [(11, 21), (31, 41), (51, 61)]:
            metadata = DataPointMetadata(key, value, update_checksum=True)
            variable.add(metadata)
            self.assertEqual(variable.metadata[str(key)], str(value))

        self.assertEqual(len(variable.metadata), 9)
        self.assertEqual(len(variable.checksum), 64)
        self.assertEqual(variable.checksum, '''\
2518ce8c9dc0683ef87a6a438c8c79c2ae3fd8ffd38032b6c1d253057d04c8f7''')


class TestTargetDataPoints(unittest.TestCase):
    """Checks all functions and methods."""

    #########################################################################
    # General object setup
    #########################################################################

    def test___init__(self):
        """Testing function __init__."""
        # Setup TargetDataPoints
        target = 'localhost'
        ddv = TargetDataPoints(target)

        # Test initial vlues
        self.assertEqual(ddv.target, target)
        self.assertFalse(ddv.valid)
        self.assertEqual(ddv.data, [])

    def test___repr__(self):
        """Testing function __repr__."""
        # Setup TargetDataPoints
        target = 'localhost'
        ddv = TargetDataPoints(target)

        # Test
        expected = ('''\
<TargetDataPoints target='localhost', valid=False, data=[]''')
        result = ddv.__repr__()
        self.assertEqual(result, expected)

    def test_add(self):
        """Testing function append."""
        # Initialize TargetDataPoints
        target = 'teddy_bear'
        ddv = TargetDataPoints(target)
        self.assertEqual(ddv.target, target)
        self.assertFalse(ddv.valid)
        self.assertEqual(ddv.data, [])

        # Setup DataPoint
        value = 457
        _key_ = 'gummy_bear'
        data_type = DATA_INT
        variable = DataPoint(_key_, value, data_type=data_type)

        # Test adding invalid value
        ddv.add(None)
        self.assertEqual(ddv.data, [])

        # Test adding variable
        ddv.add(variable)
        self.assertTrue(bool(ddv.data))
        self.assertTrue(isinstance(ddv.data, list))
        self.assertEqual(len(ddv.data), 1)
        checksum = ddv.data[0].checksum

        # Test adding duplicate variable (There should be no changes)
        ddv.add(variable)
        self.assertTrue(bool(ddv.data))
        self.assertTrue(isinstance(ddv.data, list))
        self.assertEqual(len(ddv.data), 1)
        self.assertEqual(checksum, ddv.data[0].checksum)

        # Test the values in the variable
        _variable = ddv.data[0]
        self.assertEqual(_variable.data_type, data_type)
        self.assertEqual(_variable.value, value)
        self.assertEqual(_variable.key, _key_)


class TestAgentPolledData(unittest.TestCase):
    """Checks all functions and methods."""

    #########################################################################
    # General object setup
    #########################################################################

    config = Config()

    def test___init__(self):
        """Testing function __init__."""
        # Setup AgentPolledData variable
        agent_program = 'brown_bear'
        agent_hostname = socket.getfqdn()
        polling_interval = 20
        apd = AgentPolledData(agent_program, polling_interval)
        agent_id = files.get_agent_id(agent_program, self.config)

        # Test
        self.assertTrue(bool(apd.agent_timestamp))
        self.assertEqual(
            apd.agent_polling_interval, polling_interval * 1000)
        self.assertEqual(apd.agent_id, agent_id)
        self.assertEqual(apd.agent_program, agent_program)
        self.assertEqual(apd.agent_hostname, agent_hostname)
        self.assertFalse(apd.valid)

    def test___repr__(self):
        """Testing function __repr__."""
        # Setup AgentPolledData
        agent_program = 'brown_bear'
        polling_interval = 20
        apd = AgentPolledData(agent_program, polling_interval)

        # Test
        expected = ('''\
<AgentPolledData agent_id='{}' agent_program='brown_bear', \
agent_hostname='{}', timestamp={} polling_interval={}, valid=False>\
'''.format(apd.agent_id, apd.agent_hostname, apd.agent_timestamp,
           polling_interval * 1000))
        result = apd.__repr__()
        self.assertEqual(result, expected)

    def test_add(self):
        """Testing function append."""
        # Setup AgentPolledData
        agent_program = 'panda_bear'
        polling_interval = 20
        apd = AgentPolledData(agent_program, polling_interval)

        # Initialize TargetDataPoints
        target = 'teddy_bear'
        ddv = TargetDataPoints(target)
        self.assertEqual(ddv.target, target)
        self.assertFalse(ddv.valid)
        self.assertEqual(ddv.data, [])

        # Setup DataPoint
        value = 457
        _key_ = 'gummy_bear'
        data_type = DATA_INT
        variable = DataPoint(_key_, value, data_type=data_type)

        # Add data to TargetDataPoints
        self.assertFalse(ddv.valid)
        ddv.add(variable)
        self.assertTrue(ddv.valid)

        # Test add
        self.assertFalse(apd.valid)
        apd.add(None)
        self.assertFalse(apd.valid)
        apd.add(variable)
        self.assertFalse(apd.valid)
        apd.add(ddv)
        self.assertTrue(apd.valid)

        # Test contents
        data = apd.data
        self.assertTrue(isinstance(data, list))
        self.assertEqual(len(data), 1)

        _ddv = data[0]
        self.assertTrue(isinstance(_ddv, TargetDataPoints))
        self.assertEqual(_ddv.target, target)
        self.assertTrue(_ddv.valid)
        self.assertTrue(isinstance(_ddv.data, list))
        self.assertTrue(len(_ddv.data), 1)

        data = _ddv.data
        _variable = _ddv.data[0]
        self.assertEqual(_variable.data_type, data_type)
        self.assertEqual(_variable.value, value)
        self.assertEqual(_variable.key, _key_)


class TestAgentAPIVariable(unittest.TestCase):
    """Checks all functions and methods."""

    #########################################################################
    # General object setup
    #########################################################################

    def test___init__(self):
        """Testing function __init__."""
        # Setup AgentAPIVariable
        ip_bind_port = 1234
        ip_listen_address = '1.2.3.4'

        # Test defaults
        aav = AgentAPIVariable()
        self.assertEqual(aav.ip_bind_port, 20201)
        self.assertEqual(aav.ip_listen_address, '0.0.0.0')

        # Test non-defaults
        aav = AgentAPIVariable(
            ip_bind_port=ip_bind_port, ip_listen_address=ip_listen_address)
        self.assertEqual(aav.ip_bind_port, ip_bind_port)
        self.assertEqual(aav.ip_listen_address, ip_listen_address)

    def test___repr__(self):
        """Testing function __repr__."""
        # Test defaults
        aav = AgentAPIVariable()
        expected = ('''\
<AgentAPIVariable ip_bind_port=20201, ip_listen_address='0.0.0.0'>''')
        result = aav.__repr__()
        self.assertEqual(expected, result)


class TestPollingPoint(unittest.TestCase):
    """Checks all functions and methods."""

    #########################################################################
    # General object setup
    #########################################################################

    def test___init__(self):
        """Testing function __init__."""
        # Setup PollingPoint
        address = 20
        multiplier = 6
        result = PollingPoint(address=address, multiplier=multiplier)
        self.assertEqual(result.address, address)
        self.assertEqual(result.multiplier, multiplier)

        # Test with bad multiplier
        address = 25
        multipliers = [None, False, True, 'foo']
        for multiplier in multipliers:
            result = PollingPoint(address=address, multiplier=multiplier)
            self.assertEqual(result.address, address)
            self.assertEqual(result.multiplier, 1)

    def test___repr__(self):
        """Testing function __repr__."""
        # Setup variable
        address = 20
        multiplier = 6
        item = PollingPoint(address=address, multiplier=multiplier)

        # Test
        expected = ('''\
<PollingPoint address={}, multiplier={}>'''.format(address, multiplier))
        result = item.__repr__()
        self.assertEqual(result, expected)


class TestIPTargetPollingPoints(unittest.TestCase):
    """Checks all functions and methods."""

    #########################################################################
    # General object setup
    #########################################################################

    def test___init__(self):
        """Testing function __init__."""
        # Test
        for address in [123, 'koala_bear', '1.2.3', None, True, False, {}, []]:
            result = IPTargetPollingPoints(address)
            self.assertFalse(result.valid)

        for address in ['127.0.0.1', '::1', 'localhost', 'www.google.com']:
            result = IPTargetPollingPoints(address)
            result.add(PollingPoint(1, 2))
            self.assertTrue(result.valid)


class TestTargetPollingPoints(unittest.TestCase):
    """Checks all functions and methods."""

    #########################################################################
    # General object setup
    #########################################################################

    def test___init__(self):
        """Testing function __init__."""
        # Setup TargetPollingPoints
        target = 'localhost'
        dpt = TargetPollingPoints(target)
        self.assertEqual(dpt.target, target)
        self.assertFalse(dpt.valid)

    def test___repr__(self):
        """Testing function __repr__."""
        # Setup variable
        target = 'localhost'
        item = TargetPollingPoints(target)
        result = item.__repr__()

        # Test
        expected = ('''\
<TargetPollingPoints target='localhost', valid=False, data=[]>''')
        result = item.__repr__()
        self.assertEqual(result, expected)

    def test_add(self):
        """Testing function append."""
        # Initialize TargetPollingPoints
        target = 'localhost'
        dpt = TargetPollingPoints(target)
        self.assertEqual(dpt.target, target)
        self.assertFalse(dpt.valid)

        # Add bad values
        values = [True, False, None, 'foo']
        for value in values:
            dpt.add(value)
            self.assertFalse(dpt.valid)

        # Add good values
        address = 20
        multiplier = 6
        value = PollingPoint(address=address, multiplier=multiplier)
        dpt.add(value)
        self.assertTrue(dpt.valid)
        self.assertEqual(len(dpt.data), 1)
        for item in dpt.data:
            self.assertEqual(item.address, address)
            self.assertEqual(item.multiplier, multiplier)

        # Try adding bad values and the results must be the same
        values = [True, False, None, 'foo']
        for value in values:
            dpt.add(value)
            self.assertTrue(dpt.valid)
            self.assertEqual(len(dpt.data), 1)
            item = dpt.data[0]
            self.assertEqual(item.address, address)
            self.assertEqual(item.multiplier, multiplier)


class TestBasicFunctions(unittest.TestCase):
    """Checks all functions and methods."""

    #########################################################################
    # General object setup
    #########################################################################

    def test__strip_non_printable(self):
        """Testing function _strip_non_printable."""
        pass

    def test__key_value_valid(self):
        """Testing function _key_value_valid."""
        # Test with valid values
        for _key, _value in [
                (1, 2), (0, 4), (5, 0), ('1', 2), (3, '4'), (0, 0), ('0', 1),
                (1, '0'), ('0', '0')]:
            (key, value, valid) = variables._key_value_valid(_key, _value)
            self.assertEqual(key, str(_key))
            self.assertEqual(value, _value)
            self.assertTrue(valid)

        # Test with invalid values
        for _key, _value in [
                (None, 2), (False, 4), ({5: 1}, 6),
                ('test', True), ('test', False), ('test', None),
                (True, 'test'), (False, 'test'), (None, 'test'),
                ('1', True), ('1', False), ('1', None),
                (True, '1'), (False, '1'), (None, '1'),
                (1, True), (1, False), (1, None),
                (True, 1), (False, 1), (None, 1)]:
            (key, value, valid) = variables._key_value_valid(_key, _value)
            self.assertIsNone(key)
            self.assertIsNone(value)
            self.assertFalse(valid)

        # Test with valid values
        for _key, _value in [
                (1, 2), (0, 4), (5, 0), ('1', 2), (3, '4'), (0, 0), ('0', 1),
                (1, '0'), ('0', '0')]:
            (key, value, valid) = variables._key_value_valid(
                _key, _value, metadata=True)
            self.assertEqual(key, str(_key))
            self.assertEqual(value, str(_value))
            self.assertTrue(valid)

        # Test with invalid values
        for _key, _value in [
                (None, 2), (False, 4), ({5: 1}, 6),
                ('test', True), ('test', False), ('test', None),
                (True, 'test'), (False, 'test'), (None, 'test'),
                ('1', True), ('1', False), ('1', None),
                (True, '1'), (False, '1'), (None, '1'),
                (1, True), (1, False), (1, None),
                (True, 1), (False, 1), (None, 1)]:
            (key, value, valid) = variables._key_value_valid(
                _key, _value, metadata=True)
            self.assertIsNone(key)
            self.assertIsNone(value)
            self.assertFalse(valid)


if __name__ == '__main__':
    # Make sure the environment is OK to run unittests
    UnittestConfig().create()

    # Do the unit test
    unittest.main()



#!/usr/bin/env python3
"""Pattoo files library."""

import os
import time
import sys
import json
from random import random
import subprocess

# PIP imports
import yaml

# Pattoo libraries
from pattoo_shared import log
from pattoo_shared import data


class _Directory():
    """A class for creating the names of hidden directories."""

    def __init__(self, config):
        """Initialize the class.

        Args:
            config: Config object

        Returns:
            None

        """
        # Initialize key variables
        self._root = config.daemon_directory()
        self._system_root = config.system_daemon_directory()
        self._cache = config.cache_directory()

    def pid(self):
        """Define the hidden pid directory.

        Args:
            None

        Returns:
            value: pid directory

        """
        # Return
        value = '{}{}pid'.format(self._system_root, os.sep)
        return value

    def lock(self):
        """Define the hidden lock directory.

        Args:
            None

        Returns:
            value: lock directory

        """
        # Return
        value = '{}{}lock'.format(self._system_root, os.sep)
        return value

    def agent_id(self):
        """Define the hidden agent_id directory.

        Args:
            None

        Returns:
            value: agent_id directory

        """
        # Return
        value = '{}{}agent_id'.format(self._root, os.sep)
        return value


class _File():
    """A class for creating the names of hidden files."""

    def __init__(self, config):
        """Initialize the class.

        Args:
            config: Config object

        Returns:
            None

        """
        # Initialize key variables
        self._directory = _Directory(config)

    def pid(self, prefix):
        """Define the hidden pid directory.

        Args:
            prefix: Prefix of file

        Returns:
            value: pid directory

        """
        # Return
        mkdir(self._directory.pid())
        value = '{}{}{}.pid'.format(self._directory.pid(), os.sep, prefix)
        return value

    def lock(self, prefix):
        """Define the hidden lock directory.

        Args:
            prefix: Prefix of file

        Returns:
            value: lock directory

        """
        # Return
        mkdir(self._directory.lock())
        value = '{}{}{}.lock'.format(self._directory.lock(), os.sep, prefix)
        return value

    def agent_id(self, agent_name):
        """Define the hidden agent_id directory.

        Args:
            agent_name: Agent name

        Returns:
            value: agent_id directory

        """
        # Return
        mkdir(self._directory.agent_id())
        value = ('''{}{}{}.agent_id\
'''.format(self._directory.agent_id(), os.sep, agent_name))
        return value


def read_yaml_files(config_directory):
    """Read the contents of all yaml files in a directory.

    Args:
        config_directory: Directory with configuration files

    Returns:
        config_dict: Single dict of combined yaml read from all files

    """
    # Initialize key variables
    yaml_found = False
    yaml_from_file = ''
    all_yaml_read = ''

    if os.path.isdir(config_directory) is False:
        log_message = (
            'Configuration directory "{}" '
            'doesn\'t exist!'.format(config_directory))
        log.log2die_safe(1026, log_message)

    # Cycle through list of files in directory
    for filename in os.listdir(config_directory):
        # Examine all the '.yaml' files in directory
        if filename.endswith('.yaml'):
            # YAML files found
            yaml_found = True

            # Read file and add to string
            filepath = '{}{}{}'.format(config_directory, os.sep, filename)
            yaml_from_file = read_yaml_file(filepath, as_string=True, die=True)

            # Append yaml from file to all yaml previously read
            all_yaml_read = '{}\n{}'.format(all_yaml_read, yaml_from_file)

    # Verify YAML files found in directory. We cannot use logging as it
    # requires a logfile location from the configuration directory to work
    # properly
    if yaml_found is False:
        log_message = '''\
No configuration files found in directory "{}" with ".yaml" extension.\
'''.format(config_directory)
        log.log2die_safe(1015, log_message)

    # Return
    config_dict = yaml.safe_load(all_yaml_read)
    return config_dict


def read_yaml_file(filepath, as_string=False, die=True):
    """Read the contents of a YAML file.

    Args:
        filepath: Path to file to be read
        as_string: Return a string if True
        die: Die if there is an error

    Returns:
        result: Dict of yaml read

    """
    # Initialize key variables
    if as_string is False:
        result = {}
    else:
        result = ''

    # Read file
    if filepath.endswith('.yaml'):
        try:
            with open(filepath, 'r') as file_handle:
                yaml_from_file = file_handle.read()
        except:
            _exception = sys.exc_info()
            log_message = '''\
Error reading file {}. Check permissions, existence and file syntax.\
'''.format(filepath)
            if bool(die) is True:
                log.log2die_safe_exception(1006, _exception, log_message)
            else:
                log.log2debug(1014, log_message)
                return {}

        # Get result
        if as_string is False:
            try:
                result = yaml.safe_load(yaml_from_file)
            except:
                _exception = sys.exc_info()
                log_message = '''\
Error reading file {}. Check permissions, existence and file syntax.\
'''.format(filepath)
                if bool(die) is True:
                    log.log2die_safe_exception(1001, _exception, log_message)
                else:
                    log.log2debug(1002, log_message)
                    return {}
        else:
            result = yaml_from_file

    else:
        # Die if not a YAML file
        log_message = '{} is not a YAML file.'.format(filepath)
        if bool(die) is True:
            log.log2die_safe(1065, log_message)
        else:
            log.log2debug(1005, log_message)
            if bool(as_string) is False:
                return {}
            else:
                return ''

    # Return
    return result


def read_json_files(_directory, die=True, age=0, count=None):
    """Read the contents of all JSON files in a directory.

    Args:
        _directory: Directory with JSON files
        die: Die if there is an error
        age: Minimum age of files in seconds
        count: Return first X number of sorted filenames is not None

    Returns:
        result: sorted list of tuples containing JSON read from each file and
            filepath. Sorting is important as it causes the files with the
            older timestamp names to be processed first. This allows the
            last_timestamp column to be incrementally processed versus some
            unexpected order. [(filepath, JSON), (filepath, JSON) ...]

    """
    # Initialize key variables
    json_found = False
    result = []
    processed = 0

    # Set age
    try:
        age = float(age)
    except:
        age = 0

    # Verify directory
    if os.path.isdir(_directory) is False:
        log_message = 'Directory "{}" doesn\'t exist!'.format(_directory)
        log.log2die(1009, log_message)

    # Cycle through list of files in directory
    for filename in sorted(os.listdir(_directory)):
        # Examine all the '.json' files in directory
        if filename.endswith('.json'):
            # Read file and add to tuple list
            filepath = '{}{}{}'.format(_directory, os.sep, filename)
            fileage = time.time() - os.stat(filepath).st_mtime
            if fileage > age:
                _data = read_json_file(filepath, die=die)
                if bool(_data) is True:
                    # JSON files found
                    json_found = True
                    result.append((filepath, _data))
                else:
                    # Ignore, don't update 'processed' value
                    log_message = ('''\
Error reading file {}. Ignoring.'''.format(filepath))
                    log.log2debug(1053, log_message)
                    continue

            # Stop if necessary
            processed += 1
            if bool(count) is True:
                if processed == count:
                    break

    # Verify JSON files found in directory. We cannot use logging as it
    # requires a logfile location from the configuration directory to work
    # properly
    if (json_found is False) and (bool(die) is True):
        log_message = '''\
No valid JSON files found in directory "{}" with ".json" extension.\
'''.format(_directory)
        log.log2die_safe(1060, log_message)

    # Return
    result.sort()
    return result


def read_json_file(filepath, die=True):
    """Read the contents of a YAML file.

    Args:
        filepath: Path to file to be read
        die: Die if there is an error

    Returns:
        result: Dict of JSON read

    """
    # Read file
    if filepath.endswith('.json'):
        try:
            with open(filepath, 'r') as file_handle:
                result = json.load(file_handle)
        except:
            _exception = sys.exc_info()
            log_message = ('''\
Error reading file {}. Check permissions, existence and file syntax.\
'''.format(filepath))
            if bool(die) is True:
                log.log2die_safe_exception(1012, _exception, log_message)
            else:
                log.log2debug(1013, log_message)
                return {}

    else:
        # Die if not a JSON file
        log_message = '{} is not a JSON file.'.format(filepath)
        if bool(die) is True:
            log.log2die_safe(1010, log_message)
        else:
            log.log2debug(1011, log_message)
            return {}

    # Return
    return result


def mkdir(directory):
    """Create a directory if it doesn't already exist.

    Args:
        directory: Directory name

    Returns:
        None

    """
    # Do work
    if os.path.exists(directory) is False:
        try:
            os.makedirs(directory, mode=0o775)
        except:
            log_message = (
                'Cannot create directory {}.'
                ''.format(directory))
            log.log2die(1090, log_message)

    # Fail if not a directory
    if os.path.isdir(directory) is False:
        log_message = (
            '{} is not a directory.'
            ''.format(directory))
        log.log2die(1043, log_message)


def pid_file(agent_name, config):
    """Get the pidfile for an agent.

    Args:
        agent_name: Agent name
        config: Config object

    Returns:
        result: Name of pid file

    """
    # Return
    f_obj = _File(config)
    result = f_obj.pid(agent_name)
    return result


def lock_file(agent_name, config):
    """Get the lockfile for an agent.

    Args:
        agent_name: Agent name
        config: Config object

    Returns:
        result: Name of lock file

    """
    # Return
    f_obj = _File(config)
    result = f_obj.lock(agent_name)
    return result


def agent_id_file(agent_name, config):
    """Get the agent_idfile for an agent.

    Args:
        agent_name: Agent name
        config: Config object


    Returns:
        result: Name of agent_id file

    """
    # Return
    f_obj = _File(config)
    result = f_obj.agent_id(agent_name)
    return result


def get_agent_id(agent_name, config):
    """Create a permanent UID for the agent_name.

    Args:
        agent_name: Agent name
        config: Config object

    Returns:
        agent_id: UID for agent

    """
    # Initialize key variables
    filename = agent_id_file(agent_name, config)

    # Read environment file with UID if it exists
    if os.path.isfile(filename):
        with open(filename) as f_handle:
            agent_id = f_handle.readline()
    else:
        # Create a UID and save
        agent_id = _generate_agent_id()
        with open(filename, 'w+') as env:
            env.write(str(agent_id))

    # Return
    return agent_id


def execute(command, die=True):
    """Run the command UNIX CLI command and record output.

    Args:
        command: CLI command to execute
        die: Die if errors found

    Returns:
        returncode: Return code of command execution

    """
    # Initialize key variables
    messages = []
    stdoutdata = ''.encode()
    stderrdata = ''.encode()
    returncode = 1

    # Run update_targets script
    do_command_list = list(command.split())

    # Create the subprocess object
    try:
        process = subprocess.Popen(
            do_command_list,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
        stdoutdata, stderrdata = process.communicate()
        returncode = process.returncode
    except:
        (etype, evalue, etraceback) = sys.exc_info()
        log_message = ('''\
Command failure: [Exception:{}, Exception Instance: {}, Stack Trace: {}]\
'''.format(etype, evalue, etraceback))
        log.log2warning(1052, log_message)

    # Crash if the return code is not 0
    if returncode != 0:
        # Print the Return Code header
        messages.append(
            'Return code:{}'.format(returncode)
        )

        # Print the STDOUT
        for line in stdoutdata.decode().split('\n'):
            messages.append(
                'STDOUT: {}'.format(line)
            )

        # Print the STDERR
        for line in stderrdata.decode().split('\n'):
            messages.append(
                'STDERR: {}'.format(line)
            )

        # Log message
        for log_message in messages:
            log.log2warning(1042, log_message)

        # Die if required after error found
        if bool(die) is True:
            log.log2die(1044, 'Command Failed: {}'.format(command))

    # Return
    return returncode


def _generate_agent_id():
    """Generate a UID.

    Args:
        None

    Returns:
        agent_id: the UID

    """
    # Create a UID and save
    prehash = '{}{}{}{}{}'.format(
        random(), random(), random(), random(), time.time())
    agent_id = data.hashstring(prehash)

    # Return
    return agent_id



#!/usr/bin/env python3
"""Test the encrypt module."""

# Standard imports
import unittest
import os
import sys
import shutil
import tempfile
from random import random, randint
import hashlib
from collections import defaultdict

# Try to create a working PYTHONPATH
EXEC_DIR = os.path.dirname(os.path.realpath(__file__))
ROOT_DIR = os.path.abspath(os.path.join(
    os.path.abspath(os.path.join(EXEC_DIR, os.pardir)), os.pardir))
_EXPECTED = '{0}pattoo-shared{0}tests{0}pattoo_shared_'.format(os.sep)
if EXEC_DIR.endswith(_EXPECTED) is True:
    # We need to prepend the path in case PattooShared has been installed
    # elsewhere on the system using PIP. This could corrupt expected results
    sys.path.insert(0, ROOT_DIR)
else:
    print('''This script is not installed in the "{0}" directory. Please fix.\
'''.format(_EXPECTED))
    sys.exit(2)

# Pattoo imports
from pattoo_shared import encrypt
from pattoo_shared import configuration
from pattoo_shared import files
from tests.libraries.configuration import UnittestConfig


class TestKeyRing(unittest.TestCase):
    """Test all methods of KeyRing class."""

    def test___init__(self):
        """Testing function __init__."""
        pass

    def test__import(self):
        """Testing function _import."""
        pass

    def test__export(self):
        """Testing function _export."""
        pass

    def test__generate(self):
        """Testing function _generate."""
        pass

    def test__get_public_keyid(self):
        """Testing function _get_public_keyid."""
        pass


class TestEncrypt(unittest.TestCase):
    """Test all methods of KeyRing class."""

    def setUp(self):
        """Run these steps before each test is performed."""
        # Initialize key variables
        config = configuration.BaseConfig()

        # Setup encryption instances
        self.instances = defaultdict(lambda: defaultdict(dict))

        directory = tempfile.mkdtemp()
        for item in range(2):
            agent = hashlib.md5('{}'.format(random()).encode()).hexdigest()
            email = files.get_agent_id(agent, config)
            self.instances[item]['directory'] = directory
            self.instances[item]['agent'] = agent
            self.instances[item]['email'] = email
            self.instances[item]['instance'] = encrypt.Encryption(
                agent, directory)

    def tearDown(self):
        """Run these steps after each test is performed."""
        for _, data_ in self.instances.items():
            for key, value in data_.items():
                if key == 'directory':
                    if os.path.isdir(value) is True:
                        shutil.rmtree(value)

    def test___init__(self):
        """Testing function __init__."""
        pass

    def test_decrypt(self):
        """Testing function decrypt."""
        # Initialize key Variable
        expected = '{}'.format(random())

        # Decrypting with own fingerprint
        for _, data_ in self.instances.items():
            instance = data_['instance']
            email = data_['email']

            # Encrypt with internal fingerprint
            encrypted = instance.encrypt(expected)

            # Decrypt
            result = instance.decrypt(encrypted)
            self.assertEqual(result, expected)

            # Encrypt with external fingerprint that is the same as the
            # internal instance fingerprint
            fingerprint = instance.fingerprint(email)
            fingerprint_ = instance.fingerprint()
            self.assertEqual(fingerprint, fingerprint_)
            instance.trust(fingerprint)
            encrypted = instance.encrypt(expected)

            # Decrypt
            result = instance.decrypt(encrypted)
            self.assertEqual(result, expected)

        # Decrypting with a random existing fingerprint
        for _ in range(25):
            # Chose a random instance
            offset = randint(0, len(self.instances) - 1)
            data_ = self.instances[offset]
            email = data_['email']

            # Encrypt with external fingerprint that is the same as the
            # internal instance fingerprint
            new_offset = randint(0, len(self.instances) - 1)
            data_ = self.instances[new_offset]
            instance = data_['instance']
            fingerprint = instance.fingerprint(email)
            instance.trust(fingerprint)
            encrypted = instance.encrypt(expected)

            # Decrypt
            result = instance.decrypt(encrypted)
            self.assertEqual(result, expected)

    def test_sdecrypt(self):
        """Testing function sdecrypt."""
        # Initialize key Variable
        expected = '{}'.format(random())

        # Decrypting with own fingerprint
        for _, data_ in self.instances.items():
            passphrase = '{}'.format(random())
            instance = data_['instance']

            # Encrypt with internal fingerprint
            encrypted = instance.sencrypt(expected, passphrase)

            # Decrypt
            result = instance.sdecrypt(encrypted, passphrase)
            self.assertEqual(result, expected)

    def test__decrypt(self):
        """Testing function _decrypt."""
        # Tested by test_decrypt
        pass

    def test_encrypt(self):
        """Testing function encrypt."""
        # Tested by test_decrypt
        pass

    def test_sencrypt(self):
        """Testing function sencrypt."""
        # Tested by test_sdecrypt
        pass

    def fingerpint(self):
        """Testing function fingerpint."""
        # Tested by test_decrypt
        pass

    def test_trust(self):
        """Testing function trust."""
        # Tested by test_decrypt
        pass

    def test_pdelete(self):
        """Testing function pdelete."""
        # Tested by test_pexport
        pass

    def test_pexport(self):
        """Testing function pexport."""
        # Initialize key variables
        config = configuration.BaseConfig()

        # Test
        for _, data_ in self.instances.items():
            # Get instance information
            instance = data_['instance']
            email = data_['email']
            fingerprint = instance.fingerprint(email)

            # Create a new instance
            new_agent = hashlib.md5('{}'.format(random()).encode()).hexdigest()
            new_email = files.get_agent_id(new_agent, config)
            new_directory = tempfile.mkdtemp()
            new_instance = encrypt.Encryption(new_agent, new_directory)
            new_public_key = new_instance.pexport()
            new_fingerprint = new_instance.fingerprint(new_email)

            # There should be only one public key
            public_key = instance.pexport()

            # Import public key
            instance.pimport(new_public_key)
            result = instance.pexport(new_fingerprint)

            # There should be a match
            self.assertEqual(new_public_key, result)
            self.assertNotEqual(result, public_key)

            # Test deletion of nonexistent fingerprint
            status = instance.pdelete('foo')
            self.assertFalse(status)

            # Test deletion of primary fingerprint
            status = instance.pdelete(fingerprint)
            self.assertFalse(status)

            # Test deletion of primary fingerprint
            status = instance.pdelete(new_fingerprint)
            self.assertTrue(status)
            result = instance.pexport(new_fingerprint)
            self.assertIsNone(result)

            # Delete directory
            shutil.rmtree(new_directory)

    def test_pimport(self):
        """Testing function pimport."""
        # Tested by test_pexport
        pass


class TestFunctions(unittest.TestCase):
    """Test all methods of KeyRing class."""

    def test_generate_key(self):
        """Testing function generate_key."""
        # Test
        result = encrypt.generate_key()
        self.assertTrue(isinstance(result, str))
        self.assertEqual(len(result), 70)


if __name__ == '__main__':
    # Make sure the environment is OK to run unittests
    UnittestConfig().create()

    # Do the unit test
    unittest.main()



#!/usr/bin/env python3
"""Pattoo classes that manage various URL manipulations."""

# Standard imports.
import ipaddress


def url_ip_address(ip_address):
    """Adjust address for IPv6 if necessary.

    Args:
        ip_address: IP address / hostname

    Returns:
        result: Fixed address for polling

    """
    # Initialize key variables
    result = None

    # Check IP address type
    try:
        ip_object = ipaddress.ip_address(ip_address)
    except:
        # Hostname string was presented
        result = '{}'.format(ip_address)

    if bool(result) is False:
        # Is this an IPv4 address?
        ipv4 = isinstance(ip_object, ipaddress.IPv4Address)
        if ipv4 is True:
            result = '{}'.format(ip_address)
        else:
            result = '[{}]'.format(ip_address)

    # Return result
    return result



"""Generic linux daemon base class for python 3.x."""

from __future__ import print_function
import atexit
import signal
import sys
import os
import time

# Pattoo imports
from pattoo_shared import log


class Daemon():
    """A generic daemon class.

    Usage: subclass the daemon class and override the run() method.

    Modified from http://www.jejik.com/files/examples/daemon3x.py

    """

    def __init__(self, agent):
        """Initialize the class.

        Args:
            agent: Agent object

        Returns:
            None

        """
        self.name = agent.name()
        self.pidfile = agent.pidfile_parent
        self.lockfile = agent.lockfile_parent
        self._config = agent.config

    def _daemonize(self):
        """Deamonize class. UNIX double fork mechanism.

        Args:
            None

        Returns:
            None

        """
        # Initialize key variables
        daemon_log_file = self._config.log_file_daemon()

        # Make sure that the log file is accessible.
        try:
            open(daemon_log_file, 'a').close()
        except:
            log_message = '''Cannot access daemon log file {}. Please check \
file and directory permissions.'''.format(daemon_log_file)
            log.log2die(1054, log_message)

        # Create a parent process that will manage the child
        # when the code using this class is done.
        try:
            pid = os.fork()
            if pid > 0:
                # Exit first parent
                sys.exit(0)
        except OSError as err:
            log_message = 'Daemon fork #1 failed: {}'.format(err)
            log_message = '{} - PID file: {}'.format(log_message, self.pidfile)
            log.log2die(1067, log_message)

        # Decouple from parent environment
        os.chdir('{}'.format(os.sep))
        os.setsid()
        os.umask(0)

        # Do second fork
        try:
            pid = os.fork()
            if pid > 0:

                # exit from second parent
                sys.exit(0)
        except OSError as err:
            log_message = 'Daemon fork #2 failed: {}'.format(err)
            log_message = '{} - PID file: {}'.format(log_message, self.pidfile)
            log.log2die(1072, log_message)

        # Redirect standard file descriptors, but first make sure that the
        sys.stdout.flush()
        sys.stderr.flush()
        f_handle_si = open(os.devnull, 'r')
        f_handle_so = open(daemon_log_file, 'a+')
        f_handle_se = open(daemon_log_file, 'a+')
        os.dup2(f_handle_si.fileno(), sys.stdin.fileno())
        os.dup2(f_handle_so.fileno(), sys.stdout.fileno())
        os.dup2(f_handle_se.fileno(), sys.stderr.fileno())

        # write pidfile
        atexit.register(self.delpid)
        pid = str(os.getpid())
        with open(self.pidfile, 'w+') as f_handle:
            f_handle.write('{}\n'.format(pid))

    def delpid(self):
        """Delete the PID file.

        Args:
            None

        Returns:
            None

        """
        # Delete file
        if os.path.exists(self.pidfile) is True:
            try:
                os.remove(self.pidfile)
            except:
                log_message = (
                    'PID file {} already deleted'.format(self.pidfile))
                log.log2warning(1041, log_message)

    def dellock(self):
        """Delete the lock file.

        Args:
            None

        Returns:
            None

        """
        # Delete file
        if self.lockfile is not None:
            if os.path.exists(self.lockfile) is True:
                os.remove(self.lockfile)

    def start(self):
        """Start the daemon.

        Args:
            None

        Returns:
            None

        """
        # Check for a pidfile to see if the daemon already runs
        pid = _pid(self.pidfile)

        # Die if already running
        if bool(pid) is True:
            log_message = (
                'PID file: {} already exists. Daemon already running?'
                ''.format(self.pidfile))
            log.log2die(1073, log_message)

        # Start the daemon
        self._daemonize()

        # Log success
        log_message = (
            'Daemon {} started - PID file: {}'
            ''.format(self.name, self.pidfile))
        log.log2info(1070, log_message)

        # Run code for daemon
        self.run()

    def force(self):
        """Stop the daemon by deleting the lock file first.

        Args:
            None

        Returns:
            None

        """
        # Delete lock file and stop
        self.dellock()
        self.stop()

    def stop(self):
        """Stop the daemon.

        Args:
            None

        Returns:
            None

        """
        # Check for a pidfile to see if the daemon already runs
        pid = _pid(self.pidfile)
        if bool(pid) is False:
            log_message = (
                'PID file: {} does not exist. Daemon not running?'
                ''.format(self.pidfile))
            log.log2warning(1063, log_message)
            # Not an error in a restart
            return

        # Try killing the daemon process
        try:
            os.kill(pid, signal.SIGTERM)
        except OSError as err:
            error = str(err.args)
            if error.find('No such process') > 0:
                self.delpid()
                self.dellock()
            else:
                log_message = (str(err.args))
                log_message = (
                    '{} - PID file: {}'.format(log_message, self.pidfile))
                log.log2die(1068, log_message)
        except:
            log_message = (
                'Unknown daemon "stopped" error for PID file: {}'
                ''.format(self.pidfile))
            log.log2die(1066, log_message)

        # Log success
        self.delpid()
        self.dellock()
        log_message = (
            'Daemon {} stopped - PID file: {}'
            ''.format(self.name, self.pidfile))
        log.log2info(1071, log_message)

    def restart(self):
        """Restart the daemon.

        Args:
            None

        Returns:
            None

        """
        # Restart with a wait period to make sure things shutdown smoothly
        self.stop()
        time.sleep(3)
        self.start()

    def status(self):
        """Get daemon status.

        Args:
            verbose: Print message if True

        Returns:
            result: True if the PID and PID file exists

        """
        # Determine whether pid file exists
        pid = _pid(self.pidfile)
        if bool(pid) is True:
            print('Daemon is running - {}'.format(self.name))
        else:
            print('Daemon is stopped - {}'.format(self.name))
        return bool(pid)

    def run(self):
        """You should override this method when you subclass Daemon.

        It will be called after the process has been daemonized by
        start() or restart().
        """
        # Simple comment to pass linter
        pass


class GracefulDaemon(Daemon):
    """Daemon that allows for graceful shutdown

    This daemon should allow for stop/restart commands to perform graceful
    shutdown of a given process. A graceful shutdown involves checking that
    whether a current process is running and only ending the process once the
    current process has completed its currently running task.

    """

    def __init__(self, agent):
        """Initialize the class.

        Args:
            agent: Agent object

        Returns:
            None

        """
        try:
            self.graceful_timeout = agent.config.graceful_timeout()
        except AttributeError as err:
            # Sets default GracefulDaemon shutdown timeout if not defined by the
            # agent configuration
            log_message = '''Graceful Timeout configuration not set, {}\n
            Default setting to 10s'''.format(err)
            log.log2info(1100, log_message)

            self.graceful_timeout = 10

        Daemon.__init__(self, agent)

    def __daemon_running(self):
        """Determines if daemon is running

        Daemon is running based on whether it has an associated lockfile

        Args:
            None

        Return:
            running: True if daemon is currently running or conducing a process

        """
        running = False
        if self.lockfile is not None:
            if os.path.exists(self.lockfile) is True:
                running = True

        return running

    def graceful_shutdown(self, callback):
        """Wrapper class that handles graceful_shutdown prior to using
        callaback function `fn`

        Args:
            fn: callback method

        Return:
            wrapper

        """
        def wrapper():
            """Wrapper function"""
            if self.__daemon_running():
                log_message = '''{} Lock file exists, Process still
                running'''.format(self.name)
                log.log2info(1101, log_message)

            # Continually checks if daemon is still running exits loop once
            # instance graceful_timeout limit reached
            timeout_counter = time.time()
            while True:
                # Updating timeout duration
                current_duration = time.time() - timeout_counter - 1
                time.sleep(1)

                if not self.__daemon_running() is True:
                    log_message = '''Process {} no longer
                    processing'''.format(self.name)
                    log.log2info(1103, log_message)
                    break

                if current_duration >= self.graceful_timeout:
                    log_message = '''Process {} failed to shutdown, DUE TO
                    TIMEOUT'''.format(self.name)
                    log.log2info(1104, log_message)

                    log_message = '''{}, hard shutdown in
                    progress'''.format(self.name)
                    log.log2info(1105, log_message)
                    break
            callback()
        return wrapper

    def stop(self):
        """Stop the daemon gracefully.

        Uses parent class stop method after checking that daemon is no longer
        processing data or making use of a resource.

        Args:
            None

        Returns:
            None

        """
        wrapper = self.graceful_shutdown(super(GracefulDaemon, self).stop)
        wrapper()

    def restart(self):
        """Restarts the daemon gracefully.

        Uses parent class restart method after checking that daemon is not
        processing data or making use of a resource.

        Args:
            None

        Returns:
            None

        """
        wrapper = self.graceful_shutdown(super(GracefulDaemon, self).restart)
        wrapper()


def _pid(pidfile):
    """Start the daemon.

    Args:
        pidfile: Name of file containing PID

    Returns:
        result: Value of PID

    """
    # Initialize key varialbes
    result = None

    # Check for a pidfile
    try:
        with open(pidfile, 'r') as pf_handle:
            result = int(pf_handle.read().strip())

    except IOError:
        result = None

    # Return
    return result



"""Functions for configuring pattoo components."""
# Standard imports
import os
import grp
import pwd
import getpass
from collections import defaultdict

# Import dependendices
import yaml

# Import pattoo related libraries
from pattoo_shared import files, log
from pattoo_shared.installation import shared


class _Config():
    """Installation configuration file manipulation class."""

    def __init__(self, filepath, default):
        """Initialize the class.

        Args:
            filepath: Configuration filename
            default: Default configuration

        Returns:
            None

        """
        # Read data
        self._filepath = filepath
        self._default = {} if isinstance(default, dict) is False else default

    def update(self):
        """Create configuration file.

        Args:
            None

        Returns:
            The path to the configuration file

        """
        # Get configuration
        config = self.read()
        filepath = self._filepath

        # Check validity of directories, if any
        if bool(config) is False:
            # Set default
            config = self._default

        # Create directories found in configuration
        _create_directories(config)

        # Update the configuration
        try:
            f_handle = open(filepath, 'w')
        except PermissionError:
            log.log2die(1076, '''\
Insufficient permissions for creating the file:{}'''.format(filepath))
        else:
            with f_handle:
                # Convert to dict, just in case as defaultdict
                # isn't supported by yaml.safe_dump
                yaml.safe_dump(
                    dict(config), stream=f_handle, default_flow_style=False)

    def read(self):
        """Read configuration file and replace default values.

        Args:
            None

        Returns:
            config: Dict of configuration

        """
        # Initialize key variables
        filepath = self._filepath
        default = self._default
        config = {}

        # Read config
        if os.path.isfile(filepath) is True:
            try:
                f_handle = open(filepath, 'r')
            except PermissionError:
                log.log2die_safe(1078, '''\
Insufficient permissions for reading the file:{}'''.format(filepath))
            else:
                with f_handle:
                    yaml_string = f_handle.read()
                    existing = yaml.safe_load(yaml_string)

            # Merge configurations
            config = _merge_config(default, existing)
        else:
            config = default
        return config

    def validate(self):
        """Do basic configuration validation.

        Args:
            None

        Returns:
            res

        """
        # Read configuration
        config = self.read()

        # Check main keys
        for primary in self._default.keys():
            if primary not in config:
                log_message = ('''\
Section "{}" not found in configuration file {}. Please fix.\
'''.format(primary, self._filepath))
                log.log2die_safe(1055, log_message)


def create_user(user_name, directory, shell, verbose):
    """Create user and their respective group.

    Args:
        user_name: The name and group of the user being created
        directory: The home directory of the user
        shell: The shell of the user
        verbose: A boolean value that a allows the script to run in verbose
        mode

    Returns:
        None

    """
    # Ensure user has sudo privileges
    if getpass.getuser() == 'root':

        # If the group specified does not exist, it gets created
        if not group_exists(user_name):
            shared.run_script('groupadd {0}'.format(user_name), verbose)

        # If the user specified does not exist, they get created
        if not user_exists(user_name):
            shared.run_script(
                'useradd -d {1} -s {2} -g {0} {0}'.format(
                    user_name, directory, shell), verbose)
    else:
        # Die if not root
        shared.log('You are currently not running the script as root')


def user_exists(user_name):
    """Check if the user already exists.

    Args:
        user_name: The name of the user

    Returns
        True if the user exists and False if it does not

    """
    try:
        # Gets user name
        pwd.getpwnam(user_name)
        return True
    except KeyError:
        return False


def group_exists(group_name):
    """Check if the group already exists.

    Args:
        group_name: The name of the group

    Returns
        True if the group exists and False if it does not

    """
    try:
        # Gets group name
        grp.getgrnam(group_name)
        return True
    except KeyError:
        return False


def _create_directories(config):
    """Create directories found in configuration file.

    Args:
        None

    Returns:
        None

    """
    # Iterate over config dict
    for _, value in sorted(config.items()):
        if isinstance(value, dict) is True:
            for secondary_key in value.keys():
                # Directory parameter found
                if 'directory' in secondary_key:
                    if os.sep not in value.get(secondary_key):
                        log.log2die_safe(1006,
                                         'Invalid directory {}'.format(value))

                    # Attempt to create directory
                    full_directory = os.path.expanduser(
                        value.get(secondary_key))
                    if os.path.isdir(full_directory) is False:
                        files.mkdir(full_directory)

                    # Recursively set file ownership to pattoo user and group
                    if getpass.getuser() == 'root':
                        shared.chown(full_directory)


def _merge_config(default, updates):
    """Merge two lambda dicts together.

    Args:
        default: Default dict
        updates: Dict with updates for the default

    Returns:
        result: Merged dictionary

    """
    # All components of 'result' must be dicts.
    result = defaultdict(lambda: {})

    # Merge configurations
    for key, value in default.items():
        result[key] = value
        if key in updates:
            if isinstance(value, dict):
                for _key, _value in updates[key].items():
                    result[key][_key] = _value

    # Merge the other way around
    for key, value in updates.items():
        if key not in result:
            result[key] = value

    # Convert to dict to facilitate YAML file processing
    result = dict(result)
    return result


def configure_component(component_name, config_dir, default):
    """Configure individual pattoo related components and check configuration.

    Args:
        component_name: The file name for the component being configured
        config_dir: The directory with the configuration files
        default: A dictionary containing the default configuration values

    Returns:
        None

    """
    # Initialize key variables
    filepath = os.path.join(config_dir, '{}.yaml'.format(component_name))

    # Create configuration file
    config = _Config(filepath, default)
    config.update()
    config.validate()



"""Test pattoo shared environment."""
import os
import unittest
import sys
import tempfile

# Try to create a working PYTHONPATH
EXEC_DIR = os.path.dirname(os.path.realpath(__file__))
ROOT_DIR = os.path.abspath(os.path.join(
    os.path.abspath(os.path.join(
        os.path.abspath(os.path.join(
            EXEC_DIR, os.pardir)), os.pardir)), os.pardir))
_EXPECTED = '''\
{0}pattoo-shared{0}tests{0}pattoo_shared_{0}installation'''.format(os.sep)
if EXEC_DIR.endswith(_EXPECTED) is True:
    # We need to prepend the path in case PattooShared has been installed
    # elsewhere on the system using PIP. This could corrupt expected results
    sys.path.insert(0, ROOT_DIR)
else:
    print('''This script is not installed in the "{0}" directory. Please fix.\
'''.format(_EXPECTED))
    sys.exit(2)

from tests.libraries.configuration import UnittestConfig
from pattoo_shared.installation import environment, packages, shared


def pip_helper(package):
    """Retrieve pip package without version.

    Args:
        package: The pip package parsed from the pip requirements file

    Returns:
        The package with its version removed

    """
    return package.decode().split('==')[0]


class TestVenv(unittest.TestCase):
    """Checks all functions for the Pattoo environment script."""

    def test___init__(self):
        """Testing function __init__."""
        pass

    def test_create(self):
        """Testing function create."""
        pass

    def test_activate(self):
        """Testing function activate."""
        pass

    def test_deactivate(self):
        """Testing function deactivate."""
        pass


class TestFunctions(unittest.TestCase):
    """Checks all functions for the Pattoo environment script."""

    def setUp(self):
        """Prepare each test for testing."""
        self.venv_dir = tempfile.mkdtemp()
        self.venv = environment.environment_setup(self.venv_dir)

    def tearDown(self):
        """Cleanup each test after testing."""
        self.venv.deactivate()

    def test_environment_setup(self):
        """Unittest to test the environment_setup function."""

        # Ensure that there are no packages
        pip_packages = shared.run_script('python3 -m pip freeze')[1]

        # Retrieve packages without version
        installed_packages = [
            pip_helper(package) for package in pip_packages.split()]
        result = installed_packages == []
        self.assertTrue(result)

        # Test with installing a package to the venv
        packages.install_package('matplotlib', verbose=False)
        pip_packages = shared.run_script('python3 -m pip freeze')[1]

        # Retrieve packages without version
        installed_packages = [
            pip_helper(package) for package in pip_packages.split()]
        result = 'matplotlib' in installed_packages
        self.assertTrue(result)


if __name__ == '__main__':
    # Make sure the environment is OK to run unittests
    UnittestConfig().create()

    # Do the unit test
    unittest.main()



"""Functions for configuring and installing system daemons"""
# Standard imports
from __future__ import print_function
import os
import shutil
import re
import errno
import getpass
from subprocess import check_output, call
from pathlib import Path
import yaml

# Import pattoo related libraries
from pattoo_shared.installation import shared
from pattoo_shared import log


def _filepaths(directory, full_paths=True):
    """Get the filenames in the directory.

    Args:
        directory: Directory with the files
        full_paths: Give full paths if True

    Returns:
        result: List of filenames

    """
    # Initialize key variables
    if bool(full_paths) is True:
        result = [
            os.path.join(directory, filename) for filename in os.listdir(
                directory) if os.path.isfile(
                    os.path.join(directory, filename))]
    else:
        result = [filename for filename in os.listdir(
            directory) if os.path.isfile(os.path.join(directory, filename))]

    return result


def copy_service_files(target_directory, template_dir):
    """Copy service files to target directory.

    Args:
        target_directory: Target directory the files are being copied to
        template_dir: The directory that the service files are located in

    Returns:
        destination_filepaths: List of destination filepaths

    """
    # Initialize key variables
    src_dst = {}
    destination_filepaths = []

    # Get source and destination file paths
    source_filepaths = _filepaths(template_dir)
    for filepath in source_filepaths:
        src_dst[filepath] = '{}/{}'.format(
            target_directory, os.path.basename(filepath))

    # Copy files
    for source_filepath, destination_filepath in sorted(src_dst.items()):
        shutil.copyfile(source_filepath, destination_filepath)
        destination_filepaths.append(destination_filepath)

    # Make systemd aware of the new services
    activation_command = 'systemctl daemon-reload'
    if getpass.getuser() == 'root':
        call(activation_command.split())

    # Return
    return destination_filepaths


def symlink_dir(directory):
    """Get directory in which the symlinked files are located.

    Args:
        directory: Directory with the symlinks

    Returns:
        result: Directory to which the files have symlinks

    """
    # Initialize key variables
    data_dictionary = {}
    result = None
    # Get all the filenames in the directory
    filenames = _filepaths(directory)

    # Get the name of the directory to which the files are symlinked
    for filename in filenames:
        if os.path.islink(filename) is False:
            continue
        if '/etc/systemd/system/multi-user.target.wants' not in filename:
            continue
        data_dictionary[Path(filename).resolve().absolute()] = True

    # Get the first directory in the dictionary
    for key in sorted(data_dictionary.keys()):
        if '/lib/' not in str(key):
            continue
        result = os.path.dirname(str(key))
        break
    # Die if there are no symlinks
    if bool(result) is False:
        shared.log(
            'No symlinks found in the directory: "{}"'.format(directory))
    return result


def update_environment_strings(
        filepaths, config_dir, install_dir, username, group):
    """Update the environment variables in the filepaths files.

    Args:
        filepaths: List of filepaths
        config_dir: Directory where configurations will be stored
        username: Username to run daemon
        group: Group of user to run daemon

    Returns:
        None

    """
    # Initialize key variables
    env_config_path = '^Environment="PATTOO_CONFIGDIR=(.*?)"$'
    env_user = '^User=(.*?)$'
    env_group = '^Group=(.*?)$'
    env_run = '^RuntimeDirectory=(.*?)$'

    # Do the needful
    for filepath in filepaths:
        # Read files and replace matches
        lines = []
        try:
            _fp = open(filepath, 'r')
        except PermissionError:
            log.log2die_safe(1081, '''\
Insufficient permissions for reading the file: {}.'''.format(filepath))
        except FileNotFoundError:
            log.log2die_safe(1082, '''\
The file: {}, does not exist. Ensure that the file is created.''')
        else:
            with _fp:
                line = _fp.readline()

                while line:
                    # Strip line
                    _line = line.strip()

                    # Fix the binary directory
                    _line = _line.replace('INSTALLATION_DIRECTORY', install_dir)

                    # Test PATTOO_CONFIGDIR
                    if bool(re.search(env_config_path, line)) is True:
                        _line = 'Environment="PATTOO_CONFIGDIR={}"'.format(
                            config_dir)

                    # Add RuntimeDirectory and create
                    if bool(re.search(env_run, line)) is True:
                        (run_path,
                         relative_run_path) = _get_runtime_directory(config_dir)
                        _line = 'RuntimeDirectory={}'.format(relative_run_path)
                        if getpass.getuser == 'root':
                            os.makedirs(run_path, 0o750, exist_ok=True)
                            shutil.chown(run_path, user=username, group=group)

                    # Add user
                    if bool(re.search(env_user, line)) is True:
                        _line = 'User={}'.format(username)

                    # Add group
                    if bool(re.search(env_group, line)) is True:
                        _line = 'Group={}'.format(group)

                    lines.append(_line)
                    line = _fp.readline()

        # Write new output
        try:
            _fp = open(filepath, 'w')
        except PermissionError:
            log.log2die_safe(1085, '''\
Insufficient permissions for writing/creating the file: {}'''.format(filepath))
        else:
            with _fp:
                _fp.writelines('{}\n'.format(line) for line in lines)


def _get_runtime_directory(config_directory):
    """Get the RuntimeDirectory.

    Args:
        config_dir: Configuration directory

    Returns:
        tuple: (Path, Relative Path to /var/run)

    """
    result = None
    filepath = os.path.join(config_directory, 'pattoo.yaml')
    if os.path.isfile(filepath) is False:
        shared.log('{} does not exist'.format(filepath))
    # Read yaml file
    try:
        file_handle = open(filepath, 'r')
    except PermissionError:
        log.log2die_safe(1080, '''\
Insufficient permissions for reading the file: {}.''')
    else:
        with file_handle:
            yaml_from_file = file_handle.read()
    config = yaml.safe_load(yaml_from_file)

    pattoo = config.get('pattoo')
    if bool(pattoo) is True:
        result = pattoo.get('system_daemon_directory')
    if result is None:
        shared.log('''\
"system_daemon_directory" parameter not found in the {} configuration file\
'''.format(filepath))
    _result = result.replace('/var/run/', '')
    _result = _result.replace('/run/', '')
    return (result, _result)


def preflight(config_dir, etc_dir):
    """Make sure the environment is OK.

    Args:
        config_dir: Location of the configuratiion directory
        etc_dir: Location of the systemd files

    Returns:
        None

    """
    # Make sure config_dir exists
    if os.path.isdir(config_dir) is False:
        shared.log('''\
Expected configuration directory "{}" does not exist.'''.format(config_dir))

    # Verify whether the script is being run by root or sudo user
    if bool(os.getuid()) is True:
        shared.log('''\
This script must be run as the "root" user or with "sudo" privileges''')

    # Check to see whether this is a systemd system
    try:
        check_output(['pidof', 'systemd'])
    except:
        shared.log('This is not a "systemd" system. This script should not be run.')

    # Check existence of /etc/systemd/system/multi-user.target.wants directory
    if os.path.isdir(etc_dir) is False:
        shared.log('Expected systemd directory "{}" does not exist.'.format(etc_dir))


def remove_file(file_path):
    """Safely removes files that aren't symlinks.

    Args:
        file_path: The path to the file being removed

    Returns:
        None

    """
    if os.path.islink(file_path) is False:
        # Delete files that aren't symlinks
        print('{} Is not a symlink. Removing'.format(file_path))
        try:
            os.remove(file_path)
        except PermissionError:
            log.log2die_safe(1089, '''\
Insufficient permissions for removing {}'''.format(file_path))
        except OSError as e:
            # errno.ENOENT = no such file or directory
            if e.errno != errno.ENOENT:
                raise  # re-raise exception if a different error occured
            else:
                log.log2die_safe(1092, '''\
Cannot remove {}. It does not exist'''.format(file_path))


def _check_symlinks(etc_dir, daemon):
    """Ensure the files in the etc dir are symlinks.

    Args:
        etc_dir: The directory that the symlinks are located in
        symlink_dir: The directory that the symlinks point to
        daemon: The daemon being checked

    Returns:
        None

    """
    # Initialize key variables
    symlink_path = os.path.join(etc_dir, '{}.service'.format(daemon))

    # Say what we are doing
    print('Checking if the {}.service file is a symlink '.format(daemon))
    link = os.path.islink(symlink_path)
    if link is False:
        # Attempt to remove file
        remove_file(symlink_path)
        if getpass.getuser() != 'root':
            log.log2die_safe(1086, 'Current user is not root')

        print('Creating symlink for {}'.format(daemon))
        # Create symlink if it doesn't exist
        shared.run_script('systemctl enable {}'.format(daemon))
    print('OK: Symlink is present for {}'.format(daemon))


def run_daemon(daemon_name, verbose=False):
    """Start/Restart pattoo system daemons.

    Starts daemons if they aren't running and restarts them if running

    Args:
        daemon_name: The system daemon
        verbose: A boolean value to toggle verbose output

    Returns:
        None

    """
    # Initialize key variables
    command = '''\
systemctl is-active {} daemon --quiet service-name'''.format(daemon_name)

    # Check status code of daemon to see if its running
    status = shared.run_script(command, die=False, verbose=verbose)[0]
    if status == 0:
        print('''
The {} daemon is already enabled/running, restarting daemon
'''.format(daemon_name))
        # Restart daemon if its running
        install_daemon(daemon_name, 'restart', verbose=verbose)

    # Starts daemon if its not running
    else:
        print('Enabling and starting {} daemon'.format(daemon_name))
        install_daemon(daemon_name, 'start', verbose=verbose)


def install_daemon(daemon_name, command, verbose=False):
    """Enable and start/restart respective pattoo daemons.

    Args:
        daemon_name: The name of the daemon being started
        command: The command to either start, or restart the daemon
        verbose: A boolean value to toggle verbose output

    Returns:
        None

    """
    # Enable daemon
    shared.run_script('systemctl enable {}'.format(daemon_name), verbose=verbose)
    # Start/Restart daemon
    try:
        shared.run_script('systemctl {0} {1}'.format(command, daemon_name), verbose=verbose)
    except SystemExit:
        message = '''\
Unable to {0} daemon.
Run the following command to see what could be causing the problem: \
"systemctl status {1}.service"
'''.format(command, daemon_name)
        log.log2die(1087, message)


def install(daemon_list, template_dir, installation_dir, verbose=False):
    """Installs and runs all daemons entered.

    Args:
        daemon_list: A list of the daemons to be run and installed
        template_dir: The directory the tempalte files are located in
        installation_dir: The root directory of the pattoo related project
        that the daemons will be running from
        verbose: A boolean value to toggle verbose output

    Returns:
        None

    """
    # Initialize key variables
    etc_dir = '/etc/systemd/system/multi-user.target.wants'
    config_dir = '/etc/pattoo'

    # Make sure this system supports systemd and that
    # the required directories exist
    preflight(config_dir, etc_dir)

    # Check symlink location of files in that directory
    target_directory = symlink_dir(etc_dir)

    # Copy files
    destination_filepaths = copy_service_files(target_directory, template_dir)

    # Update the environment strings
    update_environment_strings(
        destination_filepaths,
        config_dir,
        installation_dir,
        'pattoo',
        'pattoo')

    # Perform daemon reload
    shared.run_script('systemctl daemon-reload', verbose=verbose)

    # Loop through daemon list and start daemons
    print('Setting up system daemons')
    for daemon in daemon_list:
        run_daemon(daemon, verbose=verbose)
        _check_symlinks(etc_dir, daemon)

